/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.communication.jndi.tribe;

import java.util.Collections;
import java.util.Map;
import java.util.TreeMap;

import javax.naming.NamingException;

import org.objectweb.petals.communication.jndi.tribe.msg.request.FullUpdateRequest;
import org.objectweb.petals.communication.jndi.tribe.msg.request.RegistryRequest;
import org.objectweb.petals.communication.jndi.tribe.msg.response.RegistryResponse;
import org.objectweb.petals.communication.jndi.tribe.thread.SocketServerThread;
import org.objectweb.petals.communication.jndi.tribe.thread.TribeThread;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.tribe.channel.ReliableGroupChannelWithGms;
import org.objectweb.tribe.channel.tcp.TcpChannelPool;
import org.objectweb.tribe.common.Address;
import org.objectweb.tribe.common.Group;
import org.objectweb.tribe.common.GroupIdentifier;
import org.objectweb.tribe.common.IpAddress;
import org.objectweb.tribe.common.Member;
import org.objectweb.tribe.gms.GroupMembershipListener;
import org.objectweb.tribe.gms.GroupMembershipService;
import org.objectweb.tribe.gms.discovery.UdpDiscoveryService;

/**
 * The TRIBe based distributed JNDI server
 * 
 * @author Christophe HAMERLING - eBMWebSourcing
 * 
 */
public class TribeServer implements GroupMembershipListener {

    public enum Status {
        NOT_INITIALIZED, INITIALIZED, STARTED, STOPPED, UNKNOWN
    }

    private Status status;

    /**
     * The port to listen incoming JNDI requests
     */
    private int port;

    /**
     * 
     */
    private String host;

    /**
     * Group channel
     * 
     * The send method is broadcast thus we have to take care when we send a
     * message because we receive our own messages if we use the send method
     * without parameters
     */
    protected ReliableGroupChannelWithGms channel;

    /**
     * Data stored in this registry
     */
    protected Map<String, Map<String, Object>> data;

    /**
     * Discovery service
     */
    protected UdpDiscoveryService discovery;

    /**
     * Boolean to know if the server has been updated.
     */
    protected boolean fullyUpdated = false;

    /**
     * Flag to know if the channel of the server is registered in the TRIBe
     * group.
     */
    private boolean groupMember = false;

    /**
     * Group membership service
     */
    protected GroupMembershipService gms;

    /**
     * Group Identifier
     */
    protected GroupIdentifier groupId;

    /**
     * Name of the Tribe group
     */
    protected String groupName = "TRIBe";

    /**
     * Registry Util
     */
    protected RegistryUtil registryUtil;

    /**
     * Socket server thread
     */
    protected SocketServerThread socketServerThread;

    /**
     * Start time (to find who is the master). The master server will be the one
     * which have the lower start time value. It meens that it will be the
     * server that have started first.
     */
    protected long startTime;

    /**
     * Tribe thread
     */
    protected TribeThread tribeThread;

    protected IpAddress multicastIP;

    protected IpAddress replyIP;

    protected LoggingUtil log;

    public GroupMembershipService getGms() {
        return gms;
    }

    /**
     * Creates a new instance of {@link TribeServer}
     * 
     * @param port
     */
    public TribeServer(String host, int port, IpAddress multicastIP,
        IpAddress replyIP, String groupName, LoggingUtil log) {
        if (host == null || multicastIP == null || replyIP == null) {
            throw new IllegalArgumentException();
        }

        this.log = log;
        this.host = host;
        this.port = port;
        this.multicastIP = multicastIP;
        this.replyIP = replyIP;
        if (groupName != null) {
            this.groupName = groupName;
        }
        status = Status.NOT_INITIALIZED;
        startTime = System.currentTimeMillis();
    }

    /**
     * Init the data.
     */
    private void initData() {
        data = Collections
            .synchronizedMap(new TreeMap<String, Map<String, Object>>());
        data.put("/", new TreeMap<String, Object>());
    }

    /**
     * Process a request
     * 
     * @param request
     *            request to process
     * @return output Object of the operation
     * @throws NamingException
     */
    public Object processRequest(RegistryRequest request)
        throws NamingException {
        Object output = null;
        switch (request.getType()) {
            /*
             * BIND
             */
            case bind:
                registryUtil.bind(request.getArg1(), request.getArg2(), request
                    .getArg3());
                break;
            /*
             * CREATE SUBCONTEXT
             */
            case createSubcontext:
                output = registryUtil.createSubcontext(request.getArg1(),
                    request.getArg2());
                break;
            /*
             * DESTROY SUBCONTEXT
             */
            case destroySubcontext:
                registryUtil.destroySubcontext(request.getArg1(), request
                    .getArg2());
                break;
            /*
             * FULL UPDATE. Send back all the registered data.
             */
            case fullUpdate:
                output = data;
                break;
            /*
             * LIST
             */
            case list:
                output = registryUtil
                    .list(request.getArg1(), request.getArg2());
                break;
            /*
             * LIST BINDINGS
             */
            case listBindings:
                output = registryUtil.listBindings(request.getArg1(), request
                    .getArg2());
                break;
            /*
             * LOOKUP
             */
            case lookup:
                output = registryUtil.lookup(request.getArg1(), request
                    .getArg2());
                break;
            /*
             * LOOKUP LINK
             */
            case lookupLink:
                output = registryUtil.lookupLink(request.getArg1(), request
                    .getArg2());
                break;
            /*
             * REBIND
             */
            case rebind:
                registryUtil.rebind(request.getArg1(), request.getArg2(),
                    request.getArg3());
                break;
            /*
             * RENAME
             */
            case rename:
                registryUtil.rename(request.getArg1(), request.getArg2(),
                    request.getArg3());
                break;
            /*
             * UNBIND
             */
            case unbind:
                registryUtil.unbind(request.getArg1(), request.getArg2());
                break;

            default:
                break;
        }
        return output;
    }

    /**
     * Process a response
     * 
     * @param response
     *            Reponse to process
     * @return output of the corresponding operation
     */
    @SuppressWarnings("unchecked")
    public Object processResponse(RegistryResponse response) {
        Object output = null;
        switch (response.getType()) {
            /*
             * FULL UPDATE
             */
            case fullUpdate:
                if (!fullyUpdated) {
                    fullyUpdated = true;
                    log.debug("Received the distributed configuration");
                    // awake the 'full update' requestor thread
                    data = (Map<String, Map<String, Object>>) response
                        .getArg1();
                    synchronized (this) {
                        notify();
                    }
                }
                break;
            /*
             * MY STARTIME
             */
            case myStartTime:
                if ((Long) response.getArg1() > startTime) {
                    output = "i can be a master";
                }
                break;
            /*
             * EXCEPTION
             */
            case exception:
                // log.debug(this.startTime + " EXCEPTION : " +
                // response.getArg1());
                break;

            default:
                break;
        }
        return output;
    }

    /**
     * Start the JNDI server
     * 
     * @throws TribeException
     */
    public void start() throws TribeException {
        if (status != Status.INITIALIZED) {
            throw new TribeException("Can ot start the server in this state : "
                + status);
        }

        try {
            doStart();
        } catch (Exception e) {
            status = Status.UNKNOWN;
            throw new TribeException("Can not start the JNDI server", e);
        }
        status = Status.STARTED;
    }

    /**
     * 
     * @throws TribeException
     */
    public void stop() throws TribeException {
        if (status != Status.STARTED) {
            throw new TribeException("Can not stop the server in this state : "
                + status);
        }

        try {
            doStop();
        } catch (Exception e) {
            status = Status.UNKNOWN;
            throw new TribeException("Can not stop the JNDI server", e);
        }
        status = Status.STOPPED;
    }

    /**
     * 
     * @throws Exception
     */
    public void init() throws Exception {
        if (status != Status.NOT_INITIALIZED) {
            throw new TribeException(
                "Can not initialize the server in this state : " + status);
        }

        try {
            doInit();
        } catch (Exception e) {
            status = Status.UNKNOWN;
            throw new TribeException("Can not stop the JNDI server", e);
        }
        status = Status.INITIALIZED;
    }

    /**
     * 
     * @throws Exception
     */
    protected void doInit() throws Exception {
        registryUtil = new RegistryUtil(this, log);
        initData();
        discovery = new UdpDiscoveryService(this.multicastIP, this.replyIP);
        gms = new GroupMembershipService(this.replyIP, TcpChannelPool
            .getChannelPool(), discovery);
        gms.registerGroupMembershipListener(this);
        channel = new ReliableGroupChannelWithGms(gms);
        groupId = new GroupIdentifier(groupName);
    }

    /**
     * Really starts the JNDI server
     * 
     * @throws Exception
     */
    protected void doStart() throws Exception {
        // Set up the threads
        socketServerThread = new SocketServerThread(this);
        tribeThread = new TribeThread(this);

        channel.join(groupId);

        synchronized (this) {
            if (!groupMember) {
                this.wait(5000);
            }
        }
        groupMember = true;

        // Start first to listen broadcasted messages
        tribeThread.start();

        // Broadcast the 'Full Update' request to get the JNDI Map content
        if (channel.getCurrentGroup().getMembers().size() > 1) {
            RegistryRequest electRequest = new FullUpdateRequest(startTime, 0);
            try {
                channel.send(electRequest);
            } catch (Exception e) {
                log.error("Problem while sending full update request", e);
            }

            // Wait for the 'Full Update' response
            synchronized (this) {
                if (fullyUpdated == false) {
                    this.wait(10000);
                }
            }
        }

        fullyUpdated = true;

        socketServerThread.start();

        // wait that the socketServerThread is fully started
        synchronized (this) {
            if (!socketServerThread.isStarted()) {
                this.wait(10000);
            }
        }

        if (!socketServerThread.isStarted()) {
            throw new TribeException("Can not start the local JNDI server");
        }
    }

    /**
     * Really stops the JNDI server
     * 
     * @throws Exception
     */
    protected void doStop() throws Exception {

        tribeThread.shutdown();
        socketServerThread.shutdown();

        channel.close();

        tribeThread.join();
        socketServerThread.join();

        gms.stop();

        // as the discovery.kill() method close the multicast socket instead of
        // only leave it, we interrupt discovery directly.
        discovery.interrupt();

        gms = null;
        discovery = null;
        channel = null;
    }

    /* ### ACCESSORS ### */

    /**
     * 
     * @return
     */
    public ReliableGroupChannelWithGms getChannel() {
        return channel;
    }

    public Map<String, Map<String, Object>> getData() {
        return data;
    }

    /**
     * 
     * @return
     */
    public long getStartTime() {
        return startTime;
    }

    public boolean isFullyUpdated() {
        return fullyUpdated;
    }

    /**
     * Get the logger
     * 
     * @return
     */
    public LoggingUtil getLog() {
        return log;
    }

    /**
     * @return the host
     */
    public String getHost() {
        return host;
    }

    /**
     * @param host
     *            the host to set
     */
    public void setHost(String host) {
        this.host = host;
    }

    /**
     * @return the port
     */
    public int getPort() {
        return port;
    }

    /**
     * @param port
     *            the port to set
     */
    public void setPort(int port) {
        this.port = port;
    }

    /**
     * 
     * @param data
     */
    public void setData(Map<String, Map<String, Object>> data) {
        this.data = data;
    }

    /**
     * 
     * @param fullyUpdated
     */
    public void setFullyUpdated(boolean fullyUpdated) {
        this.fullyUpdated = fullyUpdated;
    }

    public void groupComposition(Group group, Address Adress) {
        // it can be the response from our own join of the group, not the final
        // group composition
        if (group.getMembers().size() > 1) {
            // gms.unregisterGroupMembershipListener(this);
            if (!groupMember) {
                synchronized (this) {
                    groupMember = true;
                    notify();
                }
            }
        }
    }

    public void joinMember(Member member, GroupIdentifier group) {
        // do nothing
    }

    public void quitMember(Member member, GroupIdentifier group) {
        // do nothing
    }

    public void failedMember(Member arg0, GroupIdentifier arg1, Member arg2) {
        // do nothing
    }

}
