/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: PackageHandler.java 154 2 mai 2006 ofabre $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.jbi.management.util;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.zip.ZipFile;

import org.objectweb.petals.PetalsException;
import org.ow2.petals.jbi.descriptor.JBIDescriptor;
import org.ow2.petals.jbi.descriptor.JBIDescriptorBuilder;
import org.ow2.petals.jbi.descriptor.JBIDescriptorConstants;
import org.ow2.petals.jbi.descriptor.JBIDescriptorException;
import org.objectweb.petals.platform.repository.RepositoryService;
import org.objectweb.petals.platform.repository.RepositoryServiceImpl.EntityType;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.ZipUtil;
import org.objectweb.util.monolog.api.Logger;

/**
 * Utility class for handling packages and package URIs
 *
 * @author Olivier Fabre - EBM Websourcing
 *
 */
public class PackageHandler {

	private LoggingUtil log;

	private RepositoryService repositorySrv;

	public PackageHandler(Logger logger, RepositoryService repositorySrv) {
		super();
		this.log = new LoggingUtil(logger);
		this.repositorySrv = repositorySrv;
	}

	/**
	 * Load the jbi descriptor from the given archive
	 *
	 * @param archiveURI
	 *            the archive URI
	 * @return the {@link JBIDescriptor} object
	 * @throws PetalsException
	 */
	public JBIDescriptor loadDescriptor(URI archiveURI) throws PetalsException {
		log.start();
		String msg;
		JBIDescriptor result = null;

		File tmpJbiXml = null;
		try {
			/*
			 * open the package Zip archive file and extract jbi.xml temp
			 */
			try {
				ZipFile zipArchive = ZipUtil.openZipFile(archiveURI);
				tmpJbiXml = ZipUtil.getEntryAsTemp(zipArchive,
				        JBIDescriptorConstants.JBI_DESCRIPTOR_RESOURCE);
				zipArchive.close();
			} catch (IOException e) {
				msg = "Error during jbi.xml extraction from package : "
						+ archiveURI.getPath();
				log.error(msg);

				throw new PetalsException(msg, e);
			}

			if (tmpJbiXml == null) {
				msg = "Invalid JBI package Zip archive. jbi.xml descriptor "
						+ "entry was not found: " + archiveURI.toString();
				log.error(msg);

				throw new PetalsException(msg);
			}

			/*
			 * use builder to parse and validate jbi.xml against the schema
			 */
			result = JBIDescriptorBuilder.getInstance().validateAndBuildDescriptor(
					tmpJbiXml.toURI(), false);

		} catch (JBIDescriptorException e) {
			throw new PetalsException(e);
		} finally {
			/*
			 * delete temporary file
			 */
			if (tmpJbiXml != null) {
				tmpJbiXml.delete();
			}
		}
		log.end();

		return result;
	}

	/**
	 * Load the jbi descriptor from the repository folder of the given entity
	 * (component, shared lib or service assembly).
	 *
	 * @param entityName
	 *            name of entity
	 * @param entityType
	 *            type of entity (component, shared lib or service assembly)
	 * @return the {@link JBIDescriptor} object
	 * @throws PetalsException
	 */
	public JBIDescriptor loadDescriptor(String entityName, EntityType entityType)
			throws PetalsException {
		log.start();

		String msg;
		JBIDescriptor result = null;

		/*
		 * extract descriptor file to parse and validate
		 */
		File tmpJbiXml = null;

		try {
			/*
			 * Retrieve jbi.xml file from repository
			 */
			switch (entityType) {
			case COMPONENT_TYPE:
				tmpJbiXml = new File(repositorySrv
						.getComponentInstallRoot(entityName),
						JBIDescriptorConstants.JBI_DESCRIPTOR_RESOURCE);
				break;
			case SA_TYPE:
				tmpJbiXml = new File(repositorySrv
						.getServiceAssemblyInstallRoot(entityName),
						JBIDescriptorConstants.JBI_DESCRIPTOR_RESOURCE);
				break;
			case SL_TYPE:
				tmpJbiXml = new File(repositorySrv
						.getSharedLibInstallRoot(entityName),
						JBIDescriptorConstants.JBI_DESCRIPTOR_RESOURCE);
				break;
			}

			if (tmpJbiXml == null) {
				msg = "jbi.xml descriptor entry was not "
						+ "found in repository for entity: " + entityName;
				log.error(msg);

				throw new PetalsException(msg);
			}

			/*
			 * use builder to parse and validate jbi.xml against the schema
			 */
			result = JBIDescriptorBuilder.getInstance().validateAndBuildDescriptor(
					tmpJbiXml.toURI(), false);

		} catch (JBIDescriptorException e) {
			msg = "Bad JBI descriptor or can not load the JBI Schema";
			log.error(msg, e);
			throw new PetalsException(msg, e);
		}
		log.end();

		return result;
	}

	/**
	 * Load the jbi descriptor from the unzipped package folder of the given
	 * entity (component, shared lib or service assembly).
	 *
	 * @param unzipRoot
	 *            unzipped package root in work directory
	 * @return the {@link JBIDescriptor} object
	 * @throws PetalsException
	 */
	public JBIDescriptor loadDescriptor(File unzipRoot) throws PetalsException {
		log.start();

		String msg;
		JBIDescriptor result = null;

		/*
		 * extract descriptor file to parse and validate
		 */
		File tmpJbiXml = new File(unzipRoot,
		        JBIDescriptorConstants.JBI_DESCRIPTOR_RESOURCE);

		try {
			if (tmpJbiXml == null) {
				msg = "jbi.xml descriptor entry was not "
						+ "found in entity package:"
						+ unzipRoot.getAbsolutePath();
				log.error(msg);

				throw new PetalsException(msg);
			}

			/*
			 * use builder to parse and validate jbi.xml against the schema
			 */
			result = JBIDescriptorBuilder.getInstance().validateAndBuildDescriptor(
					tmpJbiXml.toURI(), false);

		} catch (JBIDescriptorException e) {
			msg = "Bad JBI descriptor or can not load the JBI schema.";
			log.error(msg, e);
			throw new PetalsException(msg, e);
		}
		log.end();

		return result;
	}

	/**
	 * Process the given package Zip archive URL string and transform into a
	 * ready-to-use URI object. Additionally, checks if the point specified
	 * exist.
	 *
	 * @param archiveZipURL
	 * @param checkIfExist
	 * @return <code>URI</code> the proper package URI.
	 */
	public URI processAndGetPackageURI(String archiveZipURL,
			boolean checkIfExist) {
		log.start(archiveZipURL);

		String msg;
		URI uri = null;

		/*
		 * Create URI
		 */
		// Replace all space by %20
		archiveZipURL = archiveZipURL.replaceAll(" ", "%20");
		try {
			URL url = new URL(archiveZipURL);

			uri = url.toURI();
		} catch (MalformedURLException mue) {
			msg = "Wrong package archive location specified. URL: "
					+ archiveZipURL;

			log.error(msg, mue);

			throw new IllegalArgumentException(msg, mue);
		} catch (URISyntaxException use) {
			msg = "Syntax error converting to URI from given location. URL: "
					+ archiveZipURL;
			log.error(msg, use);

			throw new IllegalArgumentException(msg, use);
		}

		/*
		 * check if exists
		 */
		if (checkIfExist) {
			File archive = new File(uri);

			if (!archive.exists()) {
				msg = "JBI package archive not found. URL: " + archiveZipURL;
				log.error(msg);

				throw new IllegalArgumentException(msg);
			}
		}
		log.end();

		return uri;
	}

	/**
	 * Tests whether the file denoted by this given URI is a directory.
	 *
	 * @return <code>true</code> if and only if the file denoted by this URI
	 *         exists <em>and</em> is a directory; <code>false</code>
	 *         otherwise
	 *
	 * @throws SecurityException
	 *             If a security manager exists and its <code>{@link
	 *          java.lang.SecurityManager#checkRead(java.lang.String)}</code>
	 *             method denies read access to the file
	 */
	public boolean isDirectory(URI archiveURI) {
		File archive = new File(archiveURI);
		return archive.isDirectory();
	}

}
