package org.objectweb.petals.jbi.messaging.registry;

import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;

import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.petals.util.SystemUtil;

/**
 * This class is used to cache endpoint searchs. The list of all endpoints is
 * not cached because it is not used during message routing.
 * 
 * @author Christophe DENEUX - Capgemini Sud
 */
@FractalComponent
@Provides(interfaces = @Interface(name = "service", signature = EndpointCacheService.class))
public class EndpointCacheServiceImpl implements EndpointCacheService {
	
	// Default cache update period: 5min
	private static final long DEFAULT_UPDATED_PERIOD = 300000;
	
	private long cleanPeriod = DEFAULT_UPDATED_PERIOD;

	/**
	 * A cache entry containing a list of endpoints and a updated time.
	 */
	private final class EndpointCacheEntry {
		private ServiceEndpoint[] endpoints;

		private long expiredTime;

		public EndpointCacheEntry(final ServiceEndpoint[] endpoints, final long updatePeriod) {
			this.endpoints = endpoints;
			this.expiredTime = System.currentTimeMillis() + updatePeriod;
		}

		public final ServiceEndpoint[] getEndpoints() {
			return this.endpoints;
		}

		public final void setEndpoints(final ServiceEndpoint[] endpoints) {
			this.endpoints = endpoints;
		}

		public final long getExpiredTime() {
			return this.expiredTime;
		}
	}
	
	/**
	 * Cache of endpoints satisfying services.
	 */
	private ConcurrentMap<QName, EndpointCacheEntry> serviceEndpointsMap;

	/**
	 * Cache of endpoints satisfying interfaces.
	 */
	private ConcurrentMap<QName, EndpointCacheEntry> itfEndpointsMap;

	public void clear() {
		this.serviceEndpointsMap.clear();
		this.itfEndpointsMap.clear();
	}

	public ServiceEndpoint[] getInterfaceEndpoints(final QName itfName) {
		
		if (itfName == null) {
			// The list of all endpoints is not cached.
			return null;
		}

		final EndpointCacheEntry cacheEntry = this.itfEndpointsMap.get(itfName);;
		
		if (cacheEntry != null) {
			if (cacheEntry.getExpiredTime() < System.currentTimeMillis()) {
				// The cache entry is expired, we delete it and return null as
				// if no cache entry exists matching the interface name
				this.itfEndpointsMap.remove(itfName);
				return null;
			}
			else {
				return cacheEntry.getEndpoints();
			}
		}
		else {
			return null;
		}
	}

	public ServiceEndpoint[] getServiceEndpoints(final QName serviceName) {

		if (serviceName == null) {
			// The list of all endpoints is not cached.
			return null;
		}

		final EndpointCacheEntry cacheEntry = this.serviceEndpointsMap.get(serviceName);
		if (cacheEntry != null) {
			if (cacheEntry.getExpiredTime() < System.currentTimeMillis()) {
				// The cache entry is expired, we delete it and return null as
				// if no cache entry exists matching the service name
				this.serviceEndpointsMap.remove(serviceName);
				return null;
			}
			else {
				return cacheEntry.getEndpoints();
			}
		}
		else {
			return null;
		}
	}

	public void putInterfaceEndpoints(final QName itfName,
			final ServiceEndpoint[] endpoints) {
		// The list of all endpoints is not cached.
		if (itfName != null) {
			this.itfEndpointsMap.put(itfName, new EndpointCacheEntry(
					endpoints, this.cleanPeriod));
		}
		
		
	}

	public void putServiceEndpoints(final QName serviceName,
			final ServiceEndpoint[] endpoints) {
		// The list of all endpoints is not cached.
		if (serviceName != null) {
			this.serviceEndpointsMap.put(serviceName, new EndpointCacheEntry(
					endpoints, this.cleanPeriod));
		}
	}
	
	@LifeCycle(on = LifeCycleType.START)
    protected void start() throws IllegalLifeCycleException {
		
		final String cleanPeriodStr = SystemUtil.getCacheCleanPeriod();
		if (cleanPeriodStr != null) {
			this.cleanPeriod = Long.parseLong(cleanPeriodStr);
		}
		
		this.itfEndpointsMap = new ConcurrentHashMap<QName, EndpointCacheEntry>();
		this.serviceEndpointsMap = new ConcurrentHashMap<QName, EndpointCacheEntry>();

    }

    @LifeCycle(on = LifeCycleType.STOP)
    protected void stop() throws IllegalLifeCycleException {
        this.clear();
    }
}
