/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: AddressResolver.java,v 1.1 2005/07/22 10:24:27 alouis Exp $
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.jbi.messaging.routing;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Random;
import java.util.StringTokenizer;

import javax.jbi.component.Component;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.petals.communication.network.NetworkService;
import org.objectweb.petals.jbi.component.context.ComponentContextImpl;
import org.objectweb.petals.jbi.messaging.exchange.MessageExchangeImpl;
import org.objectweb.petals.jbi.messaging.registry.EndpointService;
import org.objectweb.petals.jbi.messaging.servicedesc.AbstractEndpoint;
import org.objectweb.petals.jbi.messaging.servicedesc.EndpointType;
import org.objectweb.petals.jbi.messaging.servicedesc.InternalEndpoint;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.SystemUtil;
import org.objectweb.util.monolog.api.Logger;

/**
 * The <code>AddressResolver</code> resolves implicit addresses such as
 * <code>interface</code> or <code>service</code> into an <code>endpoint</code>.
 * Rules are used to help the <code>AddressResolver</code> in the choice of the
 * <code>endpoint</code>, if more than one implement the implicit address.
 * 
 * @author Adrien LOUIS - EBM WebSourcing
 * 
 */
public class AddressResolver {

	/**
	 * the different kind of research of a endpoint: by service or by interface.
	 */
	protected static enum kindSearch {
		/**
		 * the research by service.
		 */
		SERVICE_SEARCH,

		/**
		 * the research by interface.
		 */
		INTERFACE_SEARCH
	}
	
	/**
	 * the default hardcoded strategy parameters: random,1,1,1
	 */
	public static final List<Object> DEFAULT_STRATEGY_PARAMETERS = Arrays
            .asList(new Object[] {
                    EndpointOrderer.RANDOM,
                    new Float(1f),
                    new Float(1f),
                    new Float(1f) });

	/**
	 * the targetted endpoint service.
	 */
	private EndpointService endpointService;

	/**
	 * the default configured strategy parameters.
	 */
	private static List<Object> defaultStrategyParameters;

	/**
	 * the container network service.
	 */
	private NetworkService network;

	/**
	 * this random is used to configure the load balancing strategy.
	 */
	private Random random;

	private EndpointOrderer endpointOrderer;

	/**
     * logger wrapper.
     */
    private LoggingUtil log;

    /**
     * the logger.
     */
    @Monolog(name = "logger")
    private Logger logger;

    // Static initialization
	static {

        try {
            defaultStrategyParameters = tokenizeAndAnalyseStrategy(SystemUtil
                    .getRoutingStrategy());
        } catch (final RoutingException e) {
            defaultStrategyParameters = DEFAULT_STRATEGY_PARAMETERS;
        }
        
	}

	/**
	 * Creates a new instance of {@link AddressResolver}.
	 * 
	 * @param network
	 * 		the container network
	 * @param endpointService
	 * 		the endpoint service
	 * @throws RoutingException
	 * 		impossible to instance the address resolver
	 */
	public AddressResolver(final NetworkService network,
			final EndpointService endpointService,
			final Logger logger) throws RoutingException {
	    this.logger = logger;
	    this.log = new LoggingUtil(this.logger);
		this.endpointService = endpointService;
		this.network = network;
		this.random = new Random();
		this.endpointOrderer = new EndpointOrderer(this.logger);
	}

	/**
	 * Resolve the address if necessary. Resolution consist in finding an
	 * endpoint satisfying <code>interface</code> or <code>service</code>. the
	 * source of the exchange and the exchange help the
	 * <code>AddressResolver</code> to find the endpoint. The
	 * <code>containerName</code> and <code>ComponentName</code> properties are
	 * set. If <code>containerName</code> and <code>ComponentName</code> already
	 * exist, do nothing. The check of acceptance from the two part is done here
	 * (the Component method isExchangeWithXXXOkay()).
	 * 
	 * @param source
	 * 		the component context
	 * @param exchange
	 * 		the exchange
	 * @throws RoutingException
	 * 		no destination can be found, or problem accessing the registry
	 */
	public Collection<WeightedEndpoint> resolveAddress(
			final ComponentContextImpl source,
			final MessageExchangeImpl exchange, boolean sync)
			throws RoutingException {

		// get message exchange values
		ServiceEndpoint givenEndpoint = exchange.getEndpoint();
		QName givenServiceName = exchange.getService();
		QName givenInterfaceName = exchange.getInterfaceName();

		// Construct an array of potential matching endpoints.
		// If the endpoint has been set in the messageExchange,
		// this array contains only one element, which is this endpoint.
		Collection<WeightedEndpoint> targetedEndpoints = null;

		if (givenEndpoint != null) {
			/*
			 * Case 1 : The endpoint has been specified in the message exchange,
			 * use it
			 */
			targetedEndpoints = this
					.getTargetedEndpointFromGivenEndpoint(givenEndpoint);

		} else if (givenServiceName != null) {
			/*
			 * Case 2 : The endpoint is not specified but the service name is.
			 * Get all the endpoints which are satisfying the given service
			 */
			targetedEndpoints = this.getTargetedEndpointFromGivenServiceName(
					givenServiceName, exchange, source.getComponent());

		} else if (givenInterfaceName != null) {
			/*
			 * Case 3 : Nor the endpoint nor the service have been specified in
			 * the message exchange. Get all the endpoints which resolves the
			 * given interface;
			 */
			targetedEndpoints = this.getTargetedEndpointFromGivenInterfaceName(
					givenInterfaceName, exchange, source.getComponent());

		}

		if (targetedEndpoints == null) {
			throw new RoutingException(
					"Failed to find a destination for the MessageExchange with id: "
							+ exchange.getExchangeId());
		}

		return targetedEndpoints;
	}

	/**
	 * Get the targeted endpoint from a given endpoint.
	 * 
	 * @param givenEndpoint
	 * 		the endpoint in the message exchange. May be null
	 * @return the targeted endpoint. May be null
	 * @throws RoutingException
	 * 		impossible to get the endpoint
	 */
	private List<WeightedEndpoint> getTargetedEndpointFromGivenEndpoint(
			final ServiceEndpoint givenEndpoint) throws RoutingException {

		ServiceEndpoint targetedEndpoint = null;

		if (givenEndpoint != null) {
			if (givenEndpoint instanceof AbstractEndpoint
					&& ((AbstractEndpoint) givenEndpoint).getType() == EndpointType.INTERNAL) {
				// Assume it is a real endpoint
				targetedEndpoint = givenEndpoint;
			} else {
				// Resolve the connection
				targetedEndpoint = this.endpointService.getEndpoint(
						givenEndpoint.getServiceName(), givenEndpoint
								.getEndpointName());
			}
		}

		List<WeightedEndpoint> targetedEndpoints = null;
		if (targetedEndpoint == null) {
			throw new RoutingException("The specified endpoint '"
					+ givenEndpoint + "' does not match a registered endpoint");
		} else {
			final List<ServiceEndpoint> endpoints = new ArrayList<ServiceEndpoint>(1);
			endpoints.add(targetedEndpoint);
			targetedEndpoints = EndpointOrderer.getLocalizedEndpoints(network,
					endpoints);
		}

		return targetedEndpoints;
	}

	/**
	 * Get the targeted endpoint from a given service name.
	 * 
	 * @param givenServiceName
	 * 		the given Service Name in the message exchange. May be null
	 * @param exchange
	 * 		the exchange.
	 * @param component
	 * 		the component.
	 * @return the targeted endpoint following the specific strategy used. May
	 * 	be null
	 * @throws RoutingException
	 * 		impossible to get the endpoint
	 */
	private Collection<WeightedEndpoint> getTargetedEndpointFromGivenServiceName(
			final QName givenServiceName, final MessageExchange exchange,
			final Component component) throws RoutingException {

	    Collection<WeightedEndpoint> targetedEndpoints = null;

		if (givenServiceName != null) {
			List<ServiceEndpoint> endpoints = null;
			endpoints = this.getValidEndpoints(kindSearch.SERVICE_SEARCH,
					component, givenServiceName, exchange);

			if (endpoints.size() > 1) {
				List<Object> strategyParameters = AddressResolver
						.getAndAnalyseStrategy(exchange);
				// if there is several endpoints, we order the list according to
				// weightings
				targetedEndpoints = this.endpointOrderer.orderEndpoints(
						this.network, endpoints, exchange, strategyParameters,
						this.random);
			} else {
				targetedEndpoints = EndpointOrderer.getLocalizedEndpoints(
						network, endpoints);
			}
		}

		if (targetedEndpoints == null || targetedEndpoints.size() == 0) {
			throw new RoutingException(
					"No endpoint found for the specified service '"
							+ givenServiceName + "'");
		}

		return targetedEndpoints;
	}

	/**
	 * Get the targeted endpoint from a given interface name.
	 * 
	 * @param givenInterfaceName
	 * 		the given interface Name in the message exchange. May be null
	 * @param exchange
	 * 		the exchange.
	 * @param component
	 * 		the component.
	 * @return the targeted endpoint following the specific strategy used. May
	 * 	be null
	 * @throws RoutingException
	 * 		impossible to get the endpoint
	 */
	private Collection<WeightedEndpoint> getTargetedEndpointFromGivenInterfaceName(
			final QName givenInterfaceName, final MessageExchange exchange,
			final Component component) throws RoutingException {

	    Collection<WeightedEndpoint> targetedEndpoints = null;

		if (givenInterfaceName != null) {
			List<ServiceEndpoint> endpoints = null;

			// find an Endpoint with the specified InterfaceName
			endpoints = this.getValidEndpoints(kindSearch.INTERFACE_SEARCH,
					component, givenInterfaceName, exchange);
			if (endpoints.size() > 1) {
				final List<Object> strategyParameters = AddressResolver
						.getAndAnalyseStrategy(exchange);

				// if there is several endpoints, we order the list according to
				// weightings
				targetedEndpoints = this.endpointOrderer.orderEndpoints(
						this.network, endpoints, exchange, strategyParameters,
						this.random);
			} else {
				targetedEndpoints = EndpointOrderer.getLocalizedEndpoints(
						network, endpoints);
			}
		}

		if (targetedEndpoints == null || targetedEndpoints.size() == 0) {
			throw new RoutingException(
					"No endpoint found for the specified interface '"
							+ givenInterfaceName + "'");
		}

		return targetedEndpoints;
	}

    /**
     * Get the strategy and analyses it. If the strategy is defined as message
     * property, it is used. Otherwise the default confugred strategy is used.
     * 
     * @param exchange
     *            the message exchange in which the strategy to use can be
     *            provided. Must be non-null
     * @throws RoutingException
     *             impossible to create the endpoint chooser strategy
     * @return the analyzed list of parameters
     */
	private static List<Object> getAndAnalyseStrategy(final MessageExchange exchange)
			throws RoutingException {

		if (exchange == null) {
			throw new NullPointerException("the exchange must be non null");
		}

		final String strategy = (String) exchange
				.getProperty(EndpointOrderer.PROPERTY_STRATEGY_PROTOCOLS);
		final List<Object> result;
        if (strategy == null) {
			result = AddressResolver.defaultStrategyParameters;
		} else {
			result = AddressResolver.tokenizeAndAnalyseStrategy(strategy);

		}
		return result;
	}
	
	private static List<Object> tokenizeAndAnalyseStrategy(final String strategy)
            throws RoutingException {
        final List<String> temp = new ArrayList<String>();
        final StringTokenizer st = new StringTokenizer(strategy,
                EndpointOrderer.STRATEGY_SEPARATOR);
        while (st.hasMoreTokens()) {
            temp.add(st.nextToken().trim());
        }
        return AddressResolver.analyseStrategy(temp);
    }

	/**
	 * Return only the valid endpoints into the potential endpoints.
	 * 
	 * @param type
	 * 		the type of research (by service or interface)
	 * @param consumer
	 * 		the consumer
	 * @param givenName
	 * 		the given service or interface name
	 * @param exchange
	 * 		the exchange
	 * @return the list of valid endpoints
	 */
	protected List<ServiceEndpoint> getValidEndpoints(final kindSearch type,
			final Component consumer, final QName givenName,
			final MessageExchange exchange) {
		ServiceEndpoint[] potentielEndpoints = null;
		if (type == kindSearch.SERVICE_SEARCH) {
			potentielEndpoints = this.endpointService
					.getInternalEndpointsForService(givenName);
		} else {
			potentielEndpoints = this.endpointService
					.getInternalEndpointsForInterface(givenName);
		}
		List<ServiceEndpoint> validEndpoints = new ArrayList<ServiceEndpoint>();
		for (ServiceEndpoint providerEP : potentielEndpoints) {
			if (this.controlAcceptationExchange(consumer,
					(InternalEndpoint) providerEP, exchange)) {
				validEndpoints.add(providerEP);
			}
		}
		return validEndpoints;
	}

	/**
	 * Control that the exchnage is possible. Verify if the consumer and
	 * provider are ok to realize the exchange.
	 * 
	 * @param consumer
	 * 		the consumer
	 * @param providerEP
	 * 		the provider
	 * @param exchange
	 * 		the exchange
	 * @return true if ok
	 */
	protected boolean controlAcceptationExchange(final Component consumer,
			final InternalEndpoint providerEP, final MessageExchange exchange) {
		boolean result = false;
		// test if isExchangeWithXXXOkay has to be used.
		if (SystemUtil.isExchangeValidation()) {
			// apply isExchangeWithXXXOkay
			if (consumer.isExchangeWithProviderOkay(providerEP, exchange)
					&& providerEP.getEndpointService()
							.isExchangeWithConsumerOkayForComponent(providerEP,
									exchange)) {
				result = true;
			}
		} else {
			result = true;
		}
		return result;
	}

	/**
	 * Analyse the strategy.
	 * 
	 * @param parameters
	 * 		the list of parameters to analyze
	 * @return the analyzed list of parameters to analyze
	 * @throws RoutingException
	 * 		impossible to analyze the default routing strategy.
	 */
	private static List<Object> analyseStrategy(final List<String> parameters)
			throws RoutingException {
		List<Object> result = null;
		try {
			if (parameters != null
					&& parameters.size() == EndpointOrderer.NUMBER_STRATEGY_PARAMETERS) {

				EndpointOrderer.verifStrategyParameters(parameters.get(0)
						.toLowerCase(), Float.valueOf(parameters.get(1)), Float
						.valueOf(parameters.get(2)), Float.valueOf(parameters
						.get(3)));

				result = new ArrayList<Object>();
				result.add(parameters.get(0).toLowerCase());
				result.add(Float.valueOf(parameters.get(1)));
				result.add(Float.valueOf(parameters.get(2)));
				result.add(Float.valueOf(parameters.get(3)));
			} else {
				throw new RoutingException(
						"Invalid Parameters: "
								+ "4 parameters are required to configure the routing strategy");
			}
		} catch (NumberFormatException e) {
			throw new RoutingException("Impossible to convert the "
					+ "parameters to realize the routing strategy");
		}
		return result;
	}
}
