/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.jbi.messaging.routing;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Random;
import java.util.SortedSet;
import java.util.TreeSet;

import javax.jbi.messaging.MessageExchange;
import javax.jbi.servicedesc.ServiceEndpoint;

import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.petals.PetalsException;
import org.objectweb.petals.communication.network.NetworkService;
import org.objectweb.petals.jbi.messaging.routing.Router.Localization;
import org.objectweb.petals.jbi.messaging.servicedesc.InternalEndpoint;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.SystemUtil;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * The enpoint chooser strategy.
 * 
 * @author nsalatge - eBM WebSourcing
 * 
 */
public final class EndpointOrderer {

	/**
	 * the highest strategy.
	 */
	public static final String HIGHEST = "highest";

	/**
	 * the load balancing (random strategy).
	 */
	public static final String RANDOM = "random";

	/**
	 * Constant used to specify the strategy to use.
	 */
	public static final String PROPERTY_STRATEGY_PROTOCOLS = "org.objectweb.petals.routing.strategy";

	/**
	 * Number of parameters defined in strategy.
	 */
	public static final int NUMBER_STRATEGY_PARAMETERS = 4;

	/**
	 * the strategy separator.
	 */
	public static final String STRATEGY_SEPARATOR = ",";

	/**
     * logger wrapper.
     */
    private LoggingUtil log;

    /**
     * the logger.
     */
    @Monolog(name = "logger")
    private Logger logger;

    /**
	 * Default constructor.
	 * 
	 */
	public EndpointOrderer(final Logger logger) {
		super();
		this.logger = logger;
		this.log = new LoggingUtil(this.logger);
	}

    /**
     * Elects the endpoints list of endpoints as the message recipient, with the
     * defined strategy. Ask the consumer and the provider if they are Okay for
     * the exchange.
     * 
     * @param network
     *            the container network
     * @param endpoints
     *            the list of endpoints to select an endpoint from
     * @param exchange
     *            the exchange
     * @param strategyParameters
     *            the list of parameters of the strategy. Must be non-null
     * @param r
     *            a random parameter. May be null.
     * @return the service endpoint selected or null
     * @throws RoutingException
     *             impossible to realize this strategy
     * @throws PetalsException
     * @throws PetalsException
     */
	public SortedSet<WeightedEndpoint> orderEndpoints(final NetworkService network,
			final List<ServiceEndpoint> endpoints, final MessageExchange exchange,
			final List<Object> strategyParameters, final Random r)
			throws RoutingException {

	    // The strategy parameter number must be checked only if the strategy is not the default one.
        if (!AddressResolver.DEFAULT_STRATEGY_PARAMETERS.equals(strategyParameters)
                && (strategyParameters == null || 
                    strategyParameters.size() < EndpointOrderer.NUMBER_STRATEGY_PARAMETERS)) {

            throw new RoutingException(
                    "Invalid Parameters: "
                            + "4 parameters are required to configure the routing strategy");
        }
        
        final String type;
        final float localPonderation;
        final float remoteActivePonderation;
        final float remoteInactivePonderation;
        try {
		    final Object[] parameters = strategyParameters.toArray();
		    type = ((String) parameters[0]).toLowerCase();
		    localPonderation = ((Float)parameters[1]).floatValue();
	        remoteActivePonderation = ((Float)parameters[2]).floatValue();
	        remoteInactivePonderation = ((Float)parameters[3]).floatValue();
			
		} catch (final ClassCastException e) {
			throw new RoutingException("Impossible to convert the "
					+ "parameters to realize the routing strategy");
		}

		// The strategy parameters must be checked only if the strategy is not the default one.
        if (!AddressResolver.DEFAULT_STRATEGY_PARAMETERS.equals(strategyParameters)) {
    		EndpointOrderer.verifStrategyParameters(type, localPonderation,
    				remoteActivePonderation, remoteInactivePonderation);
		}

        if (this.logger.isLoggable(BasicLevel.DEBUG)) {
            this.log.debug("Unordered endpoint list;");
            for (ServiceEndpoint endpoint : endpoints) {
                this.log.debug("\t" + endpoint.getEndpointName());
            }
        }
        
        final SortedSet<WeightedEndpoint> orderedEndpoints;
		if (type.equals(EndpointOrderer.HIGHEST)) {
			orderedEndpoints = this.loadHighestChoose(network, endpoints,
					exchange, localPonderation, remoteActivePonderation,
					remoteInactivePonderation, r);
		} else if (type.equals(EndpointOrderer.RANDOM)) {
			orderedEndpoints = this.loadRandomChoose(network, endpoints,
					exchange, localPonderation, remoteActivePonderation,
					remoteInactivePonderation, r);
		} else {
		    // This case should never occur
		    orderedEndpoints = null;
		}

        if (this.logger.isLoggable(BasicLevel.DEBUG)) {
            this.log.debug("Ordered endpoint list;");
            for (WeightedEndpoint endpoint : orderedEndpoints) {
                this.log.debug("\t" + endpoint.endpoint.getEndpointName()
                        + " " + endpoint.localization + " " + endpoint.weight);
            }
        }

		return orderedEndpoints;
	}

	/**
	 * verify the strategy parameters.
	 * 
	 * @param type
	 * 		the strategy type
	 * @param localPonderation
	 * 		the local ponderation
	 * @param remoteActivePonderation
	 * 		the remote active ponderation
	 * @param remoteInactivePonderation
	 * 		the remote inactive ponderation
	 * @throws RoutingException
	 * 		Impossible to find an endpoint with these ponderations
	 */
	public static final void verifStrategyParameters(final String type,
			final float localPonderation, final float remoteActivePonderation,
			final float remoteInactivePonderation) throws RoutingException {
		if ((!type.equals(EndpointOrderer.HIGHEST))
				&& (!type.equals(EndpointOrderer.RANDOM))) {
			throw new RoutingException("the " + type
					+ " routing strategy is unknown");
		}

		if ((localPonderation < 0) || (remoteActivePonderation < 0)
				|| (remoteInactivePonderation < 0)) {
			throw new RoutingException(
					"impossible to have a negative parameter "
							+ "in routing strategy");
		}

		if ((localPonderation == 0) && (remoteActivePonderation == 0)
				&& (remoteInactivePonderation == 0)) {
			throw new RoutingException("Impossible to find an endpoint "
					+ "with these ponderations: 0,0,0");
		}
	}

	/**
	 * The random strategy. This strategy order endpoints in function of
	 * defined ponderations. The endpoints that have the strongest ponderation
	 * can be more easely choose in comparison with the others.
	 * 
	 * @param network
	 * 		the container network. Must be non-null
	 * @param endpoints
	 * 		the list of valid endpoints. May be null
	 * @param localPonderation
	 * 		the local ponderation
	 * @param remoteActivePonderation
	 * 		the remote active ponderation
	 * @param remoteInactivePonderation
	 * 		the remote inactive ponderation
	 * @param exchange
	 * 		the exchange. Must be non-null
	 * @param r
	 * 		a random parameter. Must be non-null.
	 * @return the service endpoit selected
	 * @throws RoutingException
	 * 		impossible to realize this strategy
	 */
	private SortedSet<WeightedEndpoint> loadRandomChoose(
			final NetworkService network,
			final List<ServiceEndpoint> endpoints,
			final MessageExchange exchange, final float localWeight,
			final float remoteActiveWeight,
			final float remoteInactiveWeight, final Random r)
			throws RoutingException {
	    
	    // We create a sorted list with its comparator defining the algorithm
        // "Random"
        final SortedSet<WeightedEndpoint> orderedEndpoints = new TreeSet<WeightedEndpoint>(
                new Comparator<WeightedEndpoint>() {

                    public int compare(final WeightedEndpoint endpointOne,
                            final WeightedEndpoint endpointTwo) {

                        // Only the endpoint weight is take into account to order endpoints
                        return endpointOne.weight.compareTo(endpointTwo.weight);
                    }

                });
        
        // Finally, we add all endpoints in the list to get sorted endpoints
        this.log.debug("Apply the algorithm RANDOM");
        EndpointOrderer.fillSortedList(network, endpoints, localWeight,
                remoteActiveWeight, remoteInactiveWeight, r, orderedEndpoints);
        
        return orderedEndpoints;
	}

    /**
     * The highest strategy. This strategy creates a ordered endpoint list
     * according to the weight of their container type (local, remote active or
     * remote inactive). If two endpoints are hosted on two containers that have
     * the same type, the container weight is multiply by a randomized value.
     * 
     * @param network
     *            the container network. Must be non-null
     * @param endpoints
     *            the list of valid endpoints. May be null
     * @param localPonderation
     *            the local ponderation
     * @param remoteActivePonderation
     *            the remote active ponderation
     * @param remoteInactivePonderation
     *            the remote incative ponderation
     * @param exchange
     *            the exchange. Must be non-null
     * @return the ordered selected service endpoints according to the algorithm
     *         "Highest"
     * @throws RoutingException
     *             impossible to contact the container network
     */
	private SortedSet<WeightedEndpoint> loadHighestChoose(
			final NetworkService network,
			final List<ServiceEndpoint> endpoints,
			final MessageExchange exchange, final float localWeight,
			final float remoteActiveWeight,
			final float remoteInactiveWeight, final Random r)
			throws RoutingException {
	    
	    
        // We create an array containing weighted localization. Array index is
        // the localization (the ordinal numbers associated to the enumeration
        // constants defined in Localization)
        final Float[] weightedLocalization = { localWeight,
                remoteActiveWeight, remoteInactiveWeight };

        // We create a sorted list with its comparator defining the algorithm
        // "Highest": The first element of the list has the highest weight in
        // the highest localization.
        final SortedSet<WeightedEndpoint> orderedEndpoints = new TreeSet<WeightedEndpoint>(
                new Comparator<WeightedEndpoint>() {

                    public int compare(final WeightedEndpoint endpointOne,
                            final WeightedEndpoint endpointTwo) {

                        // The localization is the first sort criteria. If they
                        // are equal, the weight is used.
                        // Caution: the first elements of the list must be the endpoints
                        // associated to the localization with the highest weight.
                        int result = weightedLocalization[endpointTwo.localization
                                .ordinal()]
                                .compareTo(weightedLocalization[endpointOne.localization
                                        .ordinal()]);
                        if (result != 0) {
                            return result;
                        } else {
                            return endpointOne.weight
                                    .compareTo(endpointTwo.weight);
                        }
                    }

                });
        
        // Finally, we add all endpoints in the list to get sorted endpoints
        this.log.debug("Apply the algorithm HIGHEST");
        EndpointOrderer.fillSortedList(network, endpoints, localWeight,
                remoteActiveWeight, remoteInactiveWeight, r, orderedEndpoints);

        return orderedEndpoints;
        
        
	}
	
	/**
     * Fill the ordered endpoint list with the weighted endpoints defined from
     * <code>endpoints</code> from which the weight is computed according to the
     * container type weight multiply by a random value.
     * 
     * @param network
     *            the container network. Must be non-null
     * @param endpoints
     *            the list of valid endpoints. May be null
     * @param localPonderation
     *            the local ponderation
     * @param remoteActivePonderation
     *            the remote active ponderation
     * @param remoteInactivePonderation
     *            the remote incative ponderation
     * @param r
     *            a randomized value
     * @param orderedEndpoint
     *            the ordered list that will be fill by this method with
     *            <code>endpoints</code>
     * @throws RoutingException
     *             impossible to contact the container network
     */
	private static void fillSortedList(final NetworkService network,
            final List<ServiceEndpoint> endpoints, final float localWeight,
            final float remoteActiveWeight, final float remoteInactiveWeight,
            final Random r, final SortedSet<WeightedEndpoint> orderedEndpoints)
            throws RoutingException {

        try {
            for (ServiceEndpoint providerEndpoint : endpoints) {
                if (providerEndpoint instanceof InternalEndpoint) {
                    final String endpointContainerName = ((InternalEndpoint) providerEndpoint)
                            .getContainerName();

                    if (endpointContainerName.equals(SystemUtil
                            .getContainerName())) {
                        orderedEndpoints.add(new WeightedEndpoint(
                                providerEndpoint, localWeight * r.nextFloat(),
                                Localization.LOCAL));
                    } else if (network
                            .isContainerStarted(endpointContainerName)) {
                        orderedEndpoints.add(new WeightedEndpoint(
                                providerEndpoint, remoteActiveWeight
                                        * r.nextFloat(),
                                Localization.REMOTE_ACTIVE));
                    } else {
                        orderedEndpoints.add(new WeightedEndpoint(
                                providerEndpoint, remoteInactiveWeight
                                        * r.nextFloat(),
                                Localization.REMOTE_INACTIVE));
                    }
                } else {
                    // TODO identify the kind of endpoint it is to retrieve its
                    // localization
                    orderedEndpoints.add(new WeightedEndpoint(providerEndpoint,
                            remoteActiveWeight * r.nextFloat(),
                            Localization.REMOTE_ACTIVE));
                }
            }
        } catch (PetalsException e) {
            throw new RoutingException(e);
        }
    }

	/**
	 * Localize given endpoints.
	 * 
	 * @param network
	 * @param endpoints
	 * 
	 * @return a Set of localized endpoints
	 * 
	 * @throws RoutingException
	 */
	public static final List<WeightedEndpoint> getLocalizedEndpoints(
			final NetworkService network, final List<ServiceEndpoint> endpoints)
			throws RoutingException {

		try {
		    final List<WeightedEndpoint> result = new ArrayList<WeightedEndpoint>(
	                endpoints.size());

			for (ServiceEndpoint providerEndpoint : endpoints) {
				if (providerEndpoint instanceof InternalEndpoint) {
				    final String endpointContainerName = ((InternalEndpoint) providerEndpoint)
                            .getContainerName();
					if (endpointContainerName.equals(SystemUtil
                            .getContainerName())) {
						result.add(new WeightedEndpoint(providerEndpoint,
								Localization.LOCAL));
					} else if (network
                            .isContainerStarted(endpointContainerName)) {
						result.add(new WeightedEndpoint(providerEndpoint,
								Localization.REMOTE_ACTIVE));
					} else {
						result.add(new WeightedEndpoint(providerEndpoint,
								Localization.REMOTE_INACTIVE));
					}
				} else {
					// TODO identify the kind of endpoint it is to retrieve its
					// localization
					result.add(new WeightedEndpoint(providerEndpoint,
							Localization.REMOTE_ACTIVE));
				}
			}

	        return result;

		} catch (PetalsException e) {
			throw new RoutingException(e);
		}
	}	
}
