/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: Router.java,v 1.2 2006/03/17 10:24:27 alouis Exp $
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.jbi.messaging.routing;

import org.objectweb.petals.jbi.component.context.ComponentContextImpl;
import org.objectweb.petals.jbi.messaging.exchange.MessageExchangeImpl;

/**
 * Routes messages from the <code>DeliveryChannel</code> to their
 * destinations. Performs an itinerary resolution, adding for example a
 * transformation service before delivering the message to its recipient, and
 * perform an address resolution, finding the endpoint satisfying the the
 * destination information (finding an endpoint for a specified service).
 * Resolutions are based on defined rules.
 * 
 * Routes messages pushed from the <code>Transporter</code> to the correct
 * <code>DeliveryChannel</code>
 * 
 * @author Adrien LOUIS - EBM WebSourcing
 * @author rnaudin - EBM WebSourcing
 */
public interface Router {
	
	/**
	 * Localization of the targed endpoint
	 * @author rnaudin
	 *
	 */
	public static enum Localization {
		LOCAL, REMOTE_ACTIVE, REMOTE_INACTIVE
	}

    /**
     * Send the message into the NMR.
     * 
     * @param source
     *            the source
     * @param exchange
     *            the exchange
     * @param timeOut
     *            timeout of the message (in ms). 0 for no timeout
     * @throws RoutingException
     *             impossible to send the message
     */
    void send(ComponentContextImpl source, MessageExchangeImpl exchange,
        long timeOut) throws RoutingException;

    /**
     * Send the message synchroniously into the NMR.
     * 
     * @param source
     *            the source
     * @param exchange
     *            the exchange
     * @param timeOut
     *            timeout of the message (in ms). 0 for no timeout
     * 
     * @return the reponse exchange if received, or <code>null</code> if the
     *         timeout is reached
     * 
     * @throws RoutingException
     *             impossible to realize a send synchrone
     */
    MessageExchangeImpl sendSync(ComponentContextImpl source,
        MessageExchangeImpl exchange, long timeOut) throws RoutingException;

    /**
     * Try to receive a message from MNR if there is any and return it.
     * 
     * @param source
     *            The source component context
     * @param timeoutMS
     *            The timeout to receive the incomming message. If equals to -1,
     *            the timeout is infinite
     * 
     * @return The message exchange retrieve from the <code>Transporter</code>
     * 
     * @throws RoutingException
     *             impossible to receive the message
     */
    MessageExchangeImpl receive(ComponentContextImpl source, long timeoutMS)
        throws RoutingException;

    /**
     * Inform that a component is installing. The <code>Router</code> must
     * inform its <code>Transporters</code> about the component installation.
     * 
     * @param componentName
     *            The name of the component
     * 
     * @throws RoutingException
     *             impossible to add the component
     */
    void addComponent(String componentName) throws RoutingException;

    /**
     * Inform that a component is unstalling. The <code>Router</code> must
     * inform its <code>Transporters</code> about the component
     * uninstallation.
     * 
     * @param componentName
     *            The name of the component
     * 
     * @throws RoutingException
     *             impossible to remove the component
     */
    void removeComponent(String componentName) throws RoutingException;
}
