/**
 * PETALS: PETALS Services Platform
 * Copyright (C) 2005 EBM WebSourcing
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA.
 *
 * Initial developer(s): pierre.garcia@inrialpes.fr
 */

package org.objectweb.petals.jbi.messaging.transport.dream;

import java.net.UnknownHostException;
import java.util.LinkedList;
import java.util.Map;
import java.util.Queue;
import java.util.concurrent.ConcurrentHashMap;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessageExchange;

import org.objectweb.dream.Push;
import org.objectweb.dream.PushException;
import org.objectweb.dream.cache.api.CacheException;
import org.objectweb.dream.cache.lib.BasicCacheManagerAttributeController;
import org.objectweb.dream.message.Message;
import org.objectweb.dream.message.MessageManager;
import org.objectweb.dream.message.MessageManagerType;
import org.objectweb.dream.protocol.channel.TCPIPProtocolImplAttributeController;
import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalContentException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.Service;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.fractal.util.Fractal;
import org.objectweb.petals.PetalsException;
import org.objectweb.petals.communication.jndi.JNDIService;
import org.objectweb.petals.communication.network.NetworkService;
import org.objectweb.petals.jbi.messaging.exchange.MessageExchangeImpl;
import org.objectweb.petals.jbi.messaging.servicedesc.AbstractEndpoint;
import org.objectweb.petals.jbi.messaging.transport.TransportException;
import org.objectweb.petals.jbi.messaging.transport.TransportProtocol;
import org.objectweb.petals.jbi.messaging.transport.TransportProtocolListener;
import org.objectweb.petals.jbi.messaging.transport.util.DreamUtil;
import org.objectweb.petals.jbi.messaging.transport.util.TransporterUtil;
import org.objectweb.petals.kernel.configuration.LocalConfiguration;
import org.objectweb.petals.kernel.server.FractalHelper;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.petals.util.SystemUtil;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * Dream Implementation of the Transporter
 * 
 * @author pgarcia
 */
@FractalComponent
@Provides(interfaces = { @Interface(name = "in-push", signature = Push.class),
		@Interface(name = "service", signature = TransportProtocol.class) })
public class DreamWrapper implements TransportProtocol, Push {

	// -------------------------------------------------------------------------
	// Fractal Component Names
	// -------------------------------------------------------------------------

	private static final String DREAM_CHANNEL_IN = "ChannelIn";

	private static final String DREAM_CHANNEL_OUT = "ChannelOut";

	private static final String DREAM_CHANNEL_MESSAGEPASSING = "MessagePassing";

	private static final String DREAM_PROTOCOL_TCPIP = "TCPIPProtocol";

	private static final String DREAM_PROTOCOL = "Protocol";

	private static final String DREAM_LRU_CACHE = "LRUCache";

	private static final String DREAM_LRU_CACHE_MANAGER = "CacheManager";
	
	// -------------------------------------------------------------------------
	// Constants
	// -------------------------------------------------------------------------
	/**
	 * The default maximum size of the blocking queue containing message
	 * exchanges that wait, into the transported, to be processed by component.
	 */
	private static final int DEFAULT_QUEUE_SIZE = 10000;

	/**
	 * The default listen port of the Dream transporter
	 */
	// TODO: change constant scope to "private" when the configuration service
	// will support dynanmic configuration
	public static final String DEFAULT_LISTEN_PORT = "7780";

	/**
	 * Name of the parameter that defines the listen port of the Dream
	 * transporter.
	 */
	// TODO: change constant scope to "private" when the configuration service
	// will support dynanmic configuration
	public static final String PROP_NAME_LISTEN_PORT = "transporter.dream.tcpport";

	/**
	 * The default session cache size of the Dream transporter
	 */
	private static final String DEFAULT_SESSION_CACHE_SIZE = "20";

	/**
	 * Name of the parameter that defines the session cache size of the Dream
	 * transporter.
	 */
	private static final String PROP_NAME_SESSION_CACHE_SIZE = "transporter.dream.session.cache.size";

	/**
	 * The default TCP connection timeout of the Dream transporter
	 */
	private static final String DEFAULT_CONNECTION_TIMEOUT = "10000";

	/**
	 * Name of the parameter that defines the TCP connection timeout of the
	 * Dream transporter.
	 */
	private static final String PROP_NAME_CONNECTION_TIMEOUT = "transporter.dream.connection.timeout";

	/**
	 * The default TCP connection retry number of the Dream transporter
	 */
	private static final String DEFAULT_CONNECTION_RETRY = "3";

	/**
	 * Name of the parameter that defines the TCP connection retry number of the
	 * Dream transporter.
	 */
	private static final String PROP_NAME_CONNECTION_RETRY = "transporter.dream.connection.retry";


	// -------------------------------------------------------------------------
	// Fields
	// -------------------------------------------------------------------------

	/**
	 * Map used to store incoming messages for poll() call
	 */
	protected Map<String, Queue<MessageExchangeImpl>> incomingExchanges;

	/**
	 * The maximum size of a Blocking queue containing messageExchange.
	 */
	private int queueSize = DEFAULT_QUEUE_SIZE;

	// -------------------------------------------------------------------------
	// Client Interfaces
	// -------------------------------------------------------------------------

	/**
	 * Interface used to access configuration
	 */
	@Requires(name = CONFIGURATION_ITF, signature = LocalConfiguration.class)
	protected LocalConfiguration configurationService;

	/**
	 * Interface used to get host and port of message recipient
	 */
	@Requires(name = NETWORK_ITF, signature = org.objectweb.petals.communication.network.NetworkService.class)
	protected NetworkService networkService;

	/**
	 * 
	 */
	@Requires(name = JNDI_ITF, signature = org.objectweb.petals.communication.jndi.JNDIService.class)
	protected JNDIService jndiService;

	/**
	 * Interface to send messages to another JBI environment
	 */
	@Requires(name = "out-push", signature = Push.class)
	protected Push outPushItf;

	/**
	 * Interface to manage dream message
	 */
	@Requires(name = "message-manager", signature = MessageManager.class)
	protected MessageManagerType messageManagerItf;

	/**
	 * Interface used to serialize/deserialize messages.
	 */
	@Requires(name = DISPATCHER_ITF, signature = org.objectweb.petals.jbi.messaging.transport.TransportProtocolListener.class)
	protected TransportProtocolListener dispatcher;

	/**
	 * The Fractal component
	 */
	@Service
	Component dreamTransporterComponent;

	/**
	 * The fractal exposed interface of this Fractal component
	 */
	TransportProtocol dreamTransportProtocol;

	// -------------------------------------------------------------------------
	// logger fields
	// -------------------------------------------------------------------------

	/**
	 * Fractal Logger
	 */
	@Monolog(name = "logger")
	protected Logger logger;

	/**
	 * Logger Wrapper
	 */
	protected LoggingUtil log;

	/**
	 * Dream Utilitity Object : used to create Chunks
	 */
	protected DreamUtil dreamUtil;

	// --------------------------------------------------------------------------
	// TransportProtocol Interface Implementation
	// --------------------------------------------------------------------------

	/**
	 * @see org.objectweb.petals.jbi.messaging.transport.Transporter#addDestination(java.lang.String)
	 */
	public void addDestination(String componentName) throws TransportException {
		log.call("Destination : " + componentName);

		if (incomingExchanges.containsKey(componentName)) {
			throw new TransportException("The destination " + componentName
					+ " already exist");
		} else {
			incomingExchanges.put(componentName,
					new LinkedList<MessageExchangeImpl>());
		}
	}

	/**
	 * @see org.objectweb.petals.jbi.messaging.transport.Transporter#removeDestination(java.lang.String)
	 */
	public void removeDestination(String componentName)
			throws TransportException {
		log.start("Destination : " + componentName);

		if (incomingExchanges.remove(componentName) == null) {
			log.error("The destination " + componentName + " does not exist");
		}
	}

	/**
	 * This method does following action :
	 * <ol>
	 * <li>Serialize the message exchange in a dream message.</li>
	 * <li>Get the destination of the message.</li>
	 * <li>Add it to the message.</li>
	 * <li>Send it to the transport agent.</li>
	 * </ol>
	 * 
	 * @see org.objectweb.petals.jbi.messaging.transport.Transporter#send(org.objectweb.petals.jbi.messaging.exchange.MessageExchangeImpl,
	 *      java.lang.String, java.lang.String, long)
	 */
	public void send(MessageExchangeImpl messageExchange, String componentName,
			String containerName, long timeToLive) throws TransportException {
		log.start("Send message (via Dream) to container : " + containerName
				+ " / component : " + componentName);

		// clean unused messages
		messageExchange.cleanMessages();

		// Serialization of the message
		Message msg;
		try {
			// transform the JBI message into a DREAM one
			msg = (Message) dreamUtil.jbi2dream(messageExchange);
		} catch (TransportException se) {
			throw new TransportException(
					"Error in the serialization of the message :\n "
							+ se.getMessage(), se);
		}
		log.debug("jbi message serialized in a dream message");

		// add destination = (host, port)
		dreamUtil.addChunkDestination(msg, containerName);
		log.debug("destination for the container " + containerName
				+ " added in the dream message");

		// delegate the send to the transport agent
		try {
		    outPushItf.push(msg);
		} catch (PushException e) {
			throw new TransportException(
					"Error during the send of the message " + e.getMessage()
							+ "with dream transporter", e);
		}

		log.end();
	}

	/**
	 * @see org.objectweb.petals.jbi.messaging.transport.TransportProtocol#receive(java.lang.String)
	 */
	public MessageExchangeImpl receive(String componentName)
			throws TransportException {
		log.start();

		MessageExchangeImpl responseExchange = null;

		Queue<MessageExchangeImpl> queue = this.incomingExchanges
				.get(componentName);
		synchronized (queue) {
			responseExchange = queue.poll();
		}

		log.end();
		return responseExchange;
	}

	// -------------------------------------------------------------------------
	// Dream Push Interface Implementation
	// -------------------------------------------------------------------------

	/**
	 * Serialize the dream {@link Message message} into a
	 * {@link MessageExchange jbi} one, put it in the matching queue and then
	 * tell the dispatcher a new message is available.
	 * 
	 * @param dreamMessage
	 *            the incoming dream message.
	 */
	public void push(Message dreamMessage) throws PushException {
		log.start();

		// Transform the message
		MessageExchangeImpl messageExchange;
		try {
			messageExchange = dreamUtil.dream2jbi(dreamMessage);
		} catch (JBIException e) {
			throw new PushException(
					"The message couldn't have been deserialized", e);
		}

		messageManagerItf.deleteMessage(dreamMessage);

		if (TransporterUtil.getSyncMode(messageExchange, true)) {
			if (this.logger.isLoggable(BasicLevel.DEBUG)) {
    			log.debug("The message" + messageExchange.getExchangeId()
                            + " is a synchronized response for the component "
                            + getComponentName(messageExchange));
            }
		    dispatcher.syncMessageReceived(messageExchange);
		} else {
		    if (this.logger.isLoggable(BasicLevel.DEBUG)) {
		        log.debug("The message " + messageExchange.getExchangeId()
                            + " is not a synchronized response");
		    }
			// Store the message
			String componentName = getComponentName(messageExchange);

			Queue<MessageExchangeImpl> exchangeQueue = this.incomingExchanges
					.get(componentName);

			if (exchangeQueue.size() == queueSize) {
				log
						.warning("The Dream Transporter has reached its maximum capacity for the target component '"
								+ componentName
								+ "'. The message exchange is rejected");
			} else {
				synchronized (exchangeQueue) {
					exchangeQueue.add(messageExchange);
				}
				// Warn the dispatcher
				dispatcher.messageAvailable(dreamTransportProtocol,
						componentName);
			}
		}

		log.end();
	}

	// --------------------------------------------------------------------------
	// Fractal Lifecycle interface implementation
	// --------------------------------------------------------------------------

	/**
	 * Init variables with Fractal attributes value and create/bind the topic
	 * for this container.
	 * 
	 * @throws NoSuchInterfaceException
	 * 
	 * @throws Exception
	 */
	@LifeCycle(on = LifeCycleType.START)
	public void start() throws Exception {
		log = new LoggingUtil(logger);
		log.start();
		
		try {
			this.setupComposite();
			
			// get the exposed inteface "service"
			dreamTransportProtocol = (TransportProtocol) dreamTransporterComponent
					.getFcInterface("service");

			// create dream utility object
			dreamUtil = new DreamUtil(messageManagerItf, log, networkService);

			// Maps initialization
			incomingExchanges = new ConcurrentHashMap<String, Queue<MessageExchangeImpl>>(10);
		} catch (final Exception e) {
			this.log.error(e.getMessage(), e);
			throw e;
		}

		log.end();
	}

	/**
	 * Method called when the fractal component is stopped
	 * 
	 * @throws Exception
	 */
	@LifeCycle(on = LifeCycleType.STOP)
	public void stop() throws Exception {
		log.call();
	}

	// --------------------------------------------------------------------------
	// Petals Lifecycle interface implementation
	// --------------------------------------------------------------------------

	/**
	 * @see org.objectweb.petals.PetalsLifeCycle#shutdown()
	 */
	public void shutdown() throws Exception {
		log.call();
	}

	// --------------------------------------------------------------------------
	// Utility methods
	// --------------------------------------------------------------------------

	/**
	 * Get the size of the Queue
	 */
	public int getQueueSize() {
		return queueSize;
	}

	/**
	 * Set the size of the Queue
	 * 
	 * @param queueSize
	 */
	public void setQueueSize(int queueSize) {
		this.queueSize = queueSize;
	}

	/**
	 * Get the matching component name
	 * 
	 * @param jbiMsg
	 * @param componentName
	 * @return
	 * @throws TransportException
	 */
	private static final String getComponentName(MessageExchangeImpl jbiMsg) {
		String componentName = null;
		if (jbiMsg.getRole().equals(MessageExchange.Role.CONSUMER)) {
			componentName = (jbiMsg.getConsumerEndpoint()).getComponentName();
		} else if (jbiMsg.getRole().equals(MessageExchange.Role.PROVIDER)) {
			componentName = ((AbstractEndpoint) jbiMsg.getEndpoint())
					.getComponentName();
		}

		return componentName;
	}

	/**
	 * 
	 * @param rootContentController
	 * @throws PetalsException
	 * @throws NoSuchInterfaceException
	 * @throws ADLException
	 * @throws IllegalContentException
	 * @throws IllegalLifeCycleException
	 * @throws IllegalBindingException
	 */
	public void setupComposite()
			throws PetalsException {
		
		this.log.start();

		try {
			
			final FractalHelper fractalHelper = FractalHelper.getFractalHelper();
			final Component bootstrapComponent = Fractal.getBootstrapComponent();
			final ContentController boostrapCC = Fractal.getContentController(bootstrapComponent);


			
			// Set the dynamic parameters of DREAM
			// this MUST be done before the start of the DREAM composite

			// 1. retrieve the parameters
			int dreamSessionCacheSize;
			int dreamPort;
			int connectionTimeout;
			short connectionRetry;
			try {
				dreamSessionCacheSize = Integer
						.parseInt(this.configurationService.getValue(
								PROP_NAME_SESSION_CACHE_SIZE,
								DEFAULT_SESSION_CACHE_SIZE));
				this.log.info("DREAM session cache size: " + dreamSessionCacheSize);
			} catch (NumberFormatException e) {
				throw new PetalsException(
						"The number format of the DREAM session cache size specified in the server.properties file is not correct",
						e);
			}
			try {
				// TODO: Use the configuration service when it will be available to support dynamic configuration
				dreamPort = Integer.parseInt(SystemUtil.getDreamTCPPort());
				this.log.info("DREAM port: " + dreamPort);
			} catch (NumberFormatException e) {
				throw new PetalsException(
						"The number format of the DREAM port specified in the server.properties file is not correct",
						e);
			}
			try {
				connectionRetry = Short.parseShort(this.configurationService
						.getValue(PROP_NAME_CONNECTION_RETRY,
								DEFAULT_CONNECTION_RETRY));
				this.log.info("DREAM connection retry: " + connectionRetry);
			} catch (NumberFormatException e) {
				throw new PetalsException(
						"The number format of the DREAM connection retry specified in the server.properties file is not correct",
						e);
			}
			try {
				connectionTimeout = Integer.parseInt(this.configurationService
						.getValue(PROP_NAME_CONNECTION_TIMEOUT,
								DEFAULT_CONNECTION_TIMEOUT));
				this.log.info("DREAM connection timeout: " + connectionTimeout);
			} catch (NumberFormatException e) {
				throw new PetalsException(
						"The number format of the DREAM connection timeout specified in the server.properties file is not correct",
						e);
			}
			
			// 2. set the port and host of the DREAM transporter
			this.log.debug("Setting the port and host fractal attribute.");
			// Caution the name of the Dream composite fractal component is
			// FASTTRANSPORTER_COMPONENT. It is overriden by the fast
			// transporter loading.
			Component dreamTransporterComposite = fractalHelper
					.getRecursiveComponentByName(boostrapCC,
							FractalHelper.FASTTRANSPORTER_COMPONENT);
			ContentController dreamTransporterCompositeCC = Fractal
					.getContentController(dreamTransporterComposite);

			Component fractalComponent = fractalHelper.getComponentByName(
					dreamTransporterCompositeCC, DREAM_CHANNEL_IN);
			ContentController dreanChannelInFractalComponentCC = Fractal
					.getContentController(fractalComponent);

			fractalComponent = fractalHelper.getComponentByName(
					dreanChannelInFractalComponentCC, DREAM_PROTOCOL_TCPIP);
			ContentController fractalComponentCC = Fractal
					.getContentController(fractalComponent);

			fractalComponent = fractalHelper.getComponentByName(
					fractalComponentCC, DREAM_PROTOCOL);
			TCPIPProtocolImplAttributeController tCPIPProtocolAC = (TCPIPProtocolImplAttributeController) Fractal
					.getAttributeController(fractalComponent);
			tCPIPProtocolAC.setPort(dreamPort);
			tCPIPProtocolAC.setHostName(SystemUtil.getHost());

			// 3. set timeout and retry to DREAM connection
			this.log.debug("Setting the Dream connection timeout and retry fractal attribute.");
			tCPIPProtocolAC.setConnectionTimeout(connectionTimeout);
			tCPIPProtocolAC.setConnectionRetry(connectionRetry);

			fractalComponent = fractalHelper.getComponentByName(
					dreamTransporterCompositeCC, DREAM_CHANNEL_OUT);
			ContentController dreanChannelOutFractalComponentCC = Fractal
					.getContentController(fractalComponent);

			fractalComponent = fractalHelper.getComponentByName(
					dreanChannelOutFractalComponentCC, DREAM_PROTOCOL_TCPIP);
			fractalComponentCC = Fractal.getContentController(fractalComponent);

			fractalComponent = fractalHelper.getComponentByName(
					fractalComponentCC, DREAM_PROTOCOL);
			tCPIPProtocolAC = (TCPIPProtocolImplAttributeController) Fractal
					.getAttributeController(fractalComponent);
			tCPIPProtocolAC.setConnectionTimeout(connectionTimeout);
			tCPIPProtocolAC.setConnectionRetry(connectionRetry);

			// 4.set the max number of DREAM sessions
			this.log.debug("Setting the Dream session number fractal attribute.");
			fractalComponent = fractalHelper.getComponentByName(
					dreanChannelInFractalComponentCC,
					DREAM_CHANNEL_MESSAGEPASSING);
			fractalComponentCC = Fractal.getContentController(fractalComponent);

			fractalComponent = fractalHelper.getComponentByName(fractalComponentCC,
					DREAM_LRU_CACHE);
			fractalComponentCC = Fractal.getContentController(fractalComponent);

			fractalComponent = fractalHelper.getComponentByName(fractalComponentCC,
					DREAM_LRU_CACHE_MANAGER);
			BasicCacheManagerAttributeController cacheManagerAC = (BasicCacheManagerAttributeController) Fractal
					.getAttributeController(fractalComponent);
			cacheManagerAC.setCacheMaxSize(dreamSessionCacheSize);

			fractalComponent = fractalHelper.getComponentByName(
					dreanChannelOutFractalComponentCC,
					DREAM_CHANNEL_MESSAGEPASSING);
			fractalComponentCC = Fractal.getContentController(fractalComponent);

			fractalComponent = fractalHelper.getComponentByName(fractalComponentCC,
					DREAM_LRU_CACHE);
			fractalComponentCC = Fractal.getContentController(fractalComponent);

			fractalComponent = fractalHelper.getComponentByName(fractalComponentCC,
					DREAM_LRU_CACHE_MANAGER);
			cacheManagerAC = (BasicCacheManagerAttributeController) Fractal
					.getAttributeController(fractalComponent);
			cacheManagerAC.setCacheMaxSize(dreamSessionCacheSize);
			
			this.log.end();
		} catch (ADLException e) {
			e.printStackTrace();
			throw new PetalsException(
					"Error while configuring Dream transporter", e);
		} catch (NoSuchInterfaceException e) {
			e.printStackTrace();
			throw new PetalsException(
					"Error while configuring Dream transporter", e);
		} catch (SecurityException e) {
			e.printStackTrace();
			throw new PetalsException(
					"Error while configuring Dream transporter", e);
		} catch (IllegalArgumentException e) {
			e.printStackTrace();
			throw new PetalsException(
					"Error while configuring Dream transporter", e);
		} catch (UnknownHostException e) {
			e.printStackTrace();
			throw new PetalsException(
					"Error while configuring Dream transporter", e);
		} catch (CacheException e) {
			e.printStackTrace();
			throw new PetalsException(
					"Error while configuring Dream transporter", e);
		} catch (InstantiationException e) {
			e.printStackTrace();
			throw new PetalsException(
					"Error while configuring Dream transporter", e);
		}
	}

}
