/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005-2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.kernel.server;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Properties;

import javax.management.MBeanServer;
import javax.management.MBeanServerConnection;
import javax.management.ObjectName;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.ContentController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalContentException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.api.control.NameController;
import org.objectweb.fractal.util.Fractal;
import org.objectweb.petals.PetalsException;
import org.objectweb.petals.communication.jndi.joram.JoramJNDIAgentServiceImpl;
import org.objectweb.petals.communication.jndi.tribe.TribeJNDIAgentServiceImpl;
import org.objectweb.petals.communication.network.ContainerInformation;
import org.objectweb.petals.communication.network.NetworkService;
import org.objectweb.petals.jbi.management.ManagementException;
import org.objectweb.petals.jbi.management.recovery.SystemRecoveryService;
import org.objectweb.petals.jbi.messaging.transport.DispatcherImpl;
import org.objectweb.petals.jbi.messaging.transport.TransportProtocol;
import org.objectweb.petals.jbi.messaging.transport.Transporter;
import org.objectweb.petals.kernel.admin.PetalsAdminServiceInterface;
import org.objectweb.petals.util.SystemUtil;
import org.objectweb.util.monolog.Monolog;
/**
 * This class allows to start and stop the Petals Server.
 * 
 * @author rnaudin - eBM Websourcing
 * 
 */
public class PetalsServer {

	private static final String VERSION_ACTION = "getSystemInfo";

	private static final String STOP_ACTION = "stopContainer";

	private static final String SHUTDOWN_ACTION = "shutdownContainer";

	private static final String[] MONOLOG_FILE_HANDLERS = { "petalsLF",
			"joramLF", "dreamLF" };

	/**
	 * Listener called back when petals is started or stopped
	 */
	private PetalsListener petalsListener;

	/**
	 * Thread used to stop Petals in an isolate thread
	 */
	private PetalsStopThread petalsStopThread;

	/**
	 * The fractalHelper singleton
	 */
	private static FractalHelper fractalHelper;

	/**
	 * The MBeanHelper singleton
	 */
	private static MBeanHelper mbeanHelper;

	/**
	 * Petals composite
	 */
	static private Component petalsComposite = null;

	/**
	 * Petals Content Controller
	 */
	static private ContentController petalsContentController = null;

	private Locker locker;

	/**
	 * Get the content controller of petals
	 * 
	 */
	public ContentController getPetalsContentController() {
		return petalsContentController;
	}

	/**
	 * Creates a new instance of {@link PetalsServer}
	 * 
	 * @throws PetalsException
	 */
	public PetalsServer() throws PetalsException {
		this.petalsStopThread = new PetalsStopThread(this);
		this.locker = new Locker(SystemUtil.getPetalsInstallDirectory());
		try {
			fractalHelper = FractalHelper.getFractalHelper();
			mbeanHelper = MBeanHelper.getMBeanHelper();
		} catch (ADLException e) {
			throw new PetalsException("Failed to load fractal factory");
		}
	}

	/**
	 * 
	 * @return
	 */
	public PetalsStopThread getPetalsStopThread() {
		return this.petalsStopThread;
	}

	/**
	 * Initialize petals composite
	 * 
	 * @throws PetalsException
	 * 
	 */
	private void initializePetalsComposite() throws Exception {
		try {
			petalsComposite = fractalHelper
					.createNewComponent(FractalHelper.PETALS_COMPOSITE);
			petalsContentController = Fractal
					.getContentController(petalsComposite);
			
			PetalsServer.initializeJndiLayers();
			
			PetalsServer.initializeFastTransporter();
		
	        
		} catch (final ADLException e) {
			e.printStackTrace();
			String msg = "Error creating petalsComposite.";
			throw new PetalsException(msg, e);
		} catch (final NoSuchInterfaceException e) {
			System.out.println("NoSuchInterfaceException");
			e.printStackTrace();
			String msg = "Error creating petalsComposite.";
			throw new PetalsException(msg, e);
		}
	} 
			
	private static void initializeJndiLayers() throws ADLException,
			NoSuchInterfaceException, IllegalBindingException,
			IllegalLifeCycleException, IllegalContentException {

		final Component communicationComposite = fractalHelper.getRecursiveComponentByName(
                petalsContentController, FractalHelper.COMM_COMPOSITE);
		
		// Creation of the JNDIServer if activated
        final String jndiServerType = SystemUtil.getJNDIServer();
        if ("joram".equals(jndiServerType) || "tribe".equals(jndiServerType)) {
	        final Component jndiServer;
	        if ("joram".equals(jndiServerType)) {
	            // Creation of the JNDIServer JORAM
	            jndiServer = fractalHelper.createNewComponent(JoramJNDIAgentServiceImpl.class.getName());
	        } else {
	            // Creation of the JNDIServer TRIBE
                jndiServer = fractalHelper.createNewComponent(TribeJNDIAgentServiceImpl.class.getName());
	        }

            // Add the JNDI Server component in the communication composite
	        fractalHelper.addComponent(jndiServer, communicationComposite, null);
	        final NameController jndiServerNC = Fractal.getNameController(jndiServer);
            jndiServerNC.setFcName(FractalHelper.JNDIAGENT_COMPONENT);
        }
        
        // Creation of the JNDI Client service
        final String jndiClientType = SystemUtil.getJndiClient();
        if (jndiClientType != null && jndiClientType.length() > 0) {
            
            final Component networkService = fractalHelper.getRecursiveComponentByName(
                    petalsContentController, FractalHelper.CONTAINERNETWORK_COMPONENT);

            final ContentController communicationCC = Fractal.getContentController(communicationComposite);

            final BindingController communicationBC = Fractal.getBindingController(communicationComposite);
            final BindingController networkBC = Fractal.getBindingController(networkService);

            final Component defaultJndiServiceComponent = fractalHelper.getRecursiveComponentByName(
                    communicationCC, FractalHelper.JNDICLT_COMPONENT);

            // First we create the new one
            final Component jndiClient = fractalHelper.createNewComponent(jndiClientType);
            
            // Next, we unbind port of the previous new component
            communicationBC.unbindFc("jndi");
            networkBC.unbindFc("jndi");
            
            // Last we remove the default one
            communicationCC.removeFcSubComponent(defaultJndiServiceComponent);
            
            // Next, we add the new component and bind ports on it
            communicationCC.addFcSubComponent(jndiClient);
            final Interface jndiClientServiceItf = (Interface)jndiClient.getFcInterface("service");
            communicationBC.bindFc("jndi", jndiClientServiceItf);
            networkBC.bindFc("jndi", jndiClientServiceItf);
            
            // Finally we set the correct name of the new component
            final NameController jndiClientNC = Fractal.getNameController(jndiClient);
            jndiClientNC.setFcName(FractalHelper.JNDICLT_COMPONENT);
            
        }
	}
	
	private static void initializeFastTransporter() throws ADLException,
			NoSuchInterfaceException, IllegalBindingException,
			IllegalLifeCycleException, IllegalContentException {
		
		final Component transporterComposite = fractalHelper.getRecursiveComponentByName(
                petalsContentController, FractalHelper.TRANSPORTER_COMPOSITE);
		
		// Creation of the fast transporter Fractal component. Can't be null because a default value is set.
        final Component fastTransporter = fractalHelper.createNewComponent(SystemUtil.getFastTransporter());
        
        // Add the Fast transporter component in the Transporter composite, so
		// it will be automatically started when the Transporter compiste will
		// start
        fractalHelper.addComponent(fastTransporter, transporterComposite, null);
        final NameController fastTransporterNC = Fractal.getNameController(fastTransporter);
        fastTransporterNC.setFcName(FractalHelper.FASTTRANSPORTER_COMPONENT);
        
        // Next we add the fast transporter to the dispatcher transporter list
        final Interface fastTransporterServiceItf = (Interface)fastTransporter.getFcInterface("service");
        
        final Component dispatcherService = fractalHelper.getRecursiveComponentByName(
                petalsContentController, FractalHelper.DISPATCHER_COMPONENT);
        final BindingController dispatcherBC = Fractal.getBindingController(dispatcherService);
        
        dispatcherBC.bindFc(DispatcherImpl.DELEGATED_ITF_NAME + "-fast", fastTransporterServiceItf);
        
        // Next we bind client interfaces of the Fast transporter protocol composite to
		// the right interface of the transporter component (network, jndi, configuration
        // and dispatcher)
        final ContentController transporterCC = Fractal.getContentController(transporterComposite);

        // As the network component is not a sub-component of the transporter
		// composite, we link the client interface "network" of the fast
		// transporter to the client interface "network" of the transporter
		// composite.
		final BindingController fastTransporterServiceBC = Fractal
				.getBindingController(fastTransporter);
		final Interface networkItf = (Interface) transporterCC.getFcInternalInterface("network");
		fastTransporterServiceBC.bindFc(TransportProtocol.NETWORK_ITF,
				networkItf);

		// As the jndi component is not a sub-component of the transporter
		// composite, we link the client interface "jndi" of the fast
		// transporter to the client interface "jndi" of the transporter
		// composite.
		final Interface jndiServiceItf = (Interface) transporterCC.getFcInternalInterface("jndi");
		fastTransporterServiceBC.bindFc(TransportProtocol.JNDI_ITF,
				jndiServiceItf);

		// As the configuration component is not a sub-component of the
		// transporter composite, we link the client interface "confiuration" of
		// the fast transporter to the client interface "configuration" of the
		// transporter composite.
		final Interface configurationServiceItf = (Interface) transporterCC.getFcInternalInterface("configuration");
		fastTransporterServiceBC.bindFc(TransportProtocol.CONFIGURATION_ITF,
				configurationServiceItf);

		// As the dispatcher component is a sub-component of the transporter
		// composite, we bind the client interface "dispatcher" of the fast
		// transporter to the server interface "dispatcher" of the dispatcher
		// sub-component.
		final Interface dispatcherServiceItf = (Interface) dispatcherService
				.getFcInterface("dispatcher");
		fastTransporterServiceBC.bindFc(TransportProtocol.DISPATCHER_ITF,
				dispatcherServiceItf);

	}

	/**
	 * Initialize the environment for Monolog and start the Monolog factory
	 * 
	 * @throws IOException
	 */
	private void initializeMonolog() throws IOException {
		// get the loggers properties
		URL logConfUrl = this.getClass().getResource("/loggers.properties");
		Properties logProperties = new Properties();
		logProperties.load(logConfUrl.openStream());

		// configure the Monolog handlers
		configureMonologHandlers(logProperties);

		Monolog.getMonologFactory(logProperties);
	}

	/**
	 * Configure the Monolog handlers
	 * 
	 * @param logProperties
	 */
	private void configureMonologHandlers(Properties logProperties) {
		// create logs directory is not present
		File logDir = new File(SystemUtil.getPetalsInstallDirectory(), "logs");
		logDir.mkdirs();

		// get the current date as a string
		Date currentDate = new Date(System.currentTimeMillis());
		SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy-MM-dd");
		String dateString = dateFormat.format(currentDate);

		for (String handlerName : MONOLOG_FILE_HANDLERS) {
			// set the path names of the Monologs file handlers
			String handlerOutput = logProperties.getProperty("handler."
					+ handlerName + ".output");
			if (handlerOutput != null) {
				handlerOutput = handlerOutput.replace("%d", dateString);
				logProperties.setProperty("handler." + handlerName + ".output",
						new File(logDir, handlerOutput).getAbsolutePath());
			}

			// Monolog bug on the appendMode field
			String handlerAppendMode = logProperties.getProperty("handler"
					+ handlerName + "appendMode");
			if (handlerAppendMode != null
					&& (handlerAppendMode.equals("true") || handlerAppendMode
							.equals("false"))) {
				System.setProperty(handlerAppendMode, handlerAppendMode);
			}
		}
	}

	/**
	 * Start the Petals Fractal composite.
	 * 
	 * @throws NoSuchInterfaceException
	 * @throws IllegalLifeCycleException
	 * @throws ADLException
	 * @throws IllegalBindingException
	 * @throws IllegalContentException
	 * @throws PetalsException
	 */
	private void startPetalsComposite() throws NoSuchInterfaceException,
			IllegalLifeCycleException, ADLException, IllegalContentException,
			IllegalBindingException, PetalsException {

		PetalsServer.startJndiLayers();
		
		if (!fractalHelper.startComponent(petalsComposite)) {
			throw new PetalsException("Petals server already started");
		}
	}
	
	private static void startJndiLayers() throws NoSuchInterfaceException, IllegalLifeCycleException {
		// We need to start JNDI server in first
        final String jndiServerType = SystemUtil.getJNDIServer();
        if ("joram".equals(jndiServerType) || "tribe".equals(jndiServerType)) {
            final Component jndiServerComponent = fractalHelper
                    .getRecursiveComponentByName(petalsContentController,
                            FractalHelper.JNDIAGENT_COMPONENT);
            final LifeCycleController jndiServerLCC = Fractal
                    .getLifeCycleController(jndiServerComponent);
            jndiServerLCC.startFc();
        }
		
		// The JNDIService must be started before NetworkService.
        // So we force its starting because, when it is defined with the
        // non-default value, it will be started after the NEtworkService
        // TODO: The fractal composition must be defined dynamically instead of
        // statically
        final Component jndiServiceComponent = fractalHelper
                .getRecursiveComponentByName(petalsContentController,
                        FractalHelper.JNDICLT_COMPONENT);
        final LifeCycleController jndiServiceLCC = Fractal
                .getLifeCycleController(jndiServiceComponent);
        jndiServiceLCC.startFc();
	}

	/**
	 * Stop the Petals Fractal composite.
	 * 
	 * @throws Exception
	 */
	private void stopPetalsComposite() throws Exception {

		LifeCycleController lifeCycleController = Fractal
				.getLifeCycleController(petalsComposite);

		if (lifeCycleController.getFcState()
				.equals(LifeCycleController.STARTED)) {

			// First prepare the stop of the Dispatcher, to stop the ongoing
			// transfers
			Component dispatcherComponent = fractalHelper
					.getRecursiveComponentByName(petalsContentController,
							FractalHelper.DISPATCHER_COMPONENT);
			Transporter dispatcher = (Transporter) dispatcherComponent
					.getFcInterface("service");
			dispatcher.prepareStop();

			// Then stop the Transport Protocols, to avoid blocking calls to the
			// interface
			// TransportProtocolListener.
			Component joramComponent = fractalHelper
					.getRecursiveComponentByName(petalsContentController,
							FractalHelper.RELIABLETRANSPORTER_COMPONENT);
			LifeCycleController joramLCController = Fractal
					.getLifeCycleController(joramComponent);
			joramLCController.stopFc();

			Component localComponent = fractalHelper
					.getRecursiveComponentByName(petalsContentController,
							FractalHelper.LOCALTRANSPORTER_COMPONENT);
			LifeCycleController localLCController = Fractal
					.getLifeCycleController(localComponent);
			localLCController.stopFc();

			Component dreamComponent = fractalHelper
					.getRecursiveComponentByName(petalsContentController,
							FractalHelper.FASTTRANSPORTER_COMPONENT);
			LifeCycleController dreamLCController = Fractal
					.getLifeCycleController(dreamComponent);
			dreamLCController.stopFc();

			// Stop the discovery service
			Component discoveryComponent = fractalHelper
					.getRecursiveComponentByName(petalsContentController,
							FractalHelper.DISCOVERY_COMPONENT);
			LifeCycleController discoveryLCController = Fractal
					.getLifeCycleController(discoveryComponent);
			discoveryLCController.stopFc();

			// Finally, stop the others components
			lifeCycleController.stopFc();
		}
	}

	/**
	 * Register to the Petals Admin service this Petals server instance
	 * 
	 * @throws NoSuchInterfaceException
	 * @throws PetalsException
	 */
	private void registerPetalsServer() throws NoSuchInterfaceException,
			PetalsException {
		Component petalsAdminComponent = fractalHelper
				.getRecursiveComponentByName(petalsContentController,
						FractalHelper.PETALSADMIN_COMPONENT);
		PetalsAdminServiceInterface petalsAdminService = (PetalsAdminServiceInterface) petalsAdminComponent
				.getFcInterface("service");

		petalsAdminService.setPetalsServer(this);
	}

	/**
	 * Recover the system state. All state elements are recovered (components,
	 * share libraries, service assemblies)
	 * 
	 * @throws NoSuchInterfaceException
	 * @throws ManagementException
	 * @throws ADLException
	 * @throws PetalsException
	 */
	private void recoverSystem() throws NoSuchInterfaceException,
			ManagementException, ADLException, PetalsException {
		final Component systemRecoveryComponent = fractalHelper
				.getRecursiveComponentByName(petalsContentController,
						FractalHelper.SYSTEMRECOVERY_COMPONENT);
		final SystemRecoveryService systemRecoveryService = (SystemRecoveryService) systemRecoveryComponent
				.getFcInterface("service");

		systemRecoveryService.recoverAllEntities();
		
		// When all entities are recovered, the component Fractal is no more
        // needed, so we can stop it
		try {
            final LifeCycleController systemRecoveryLCC = Fractal
                    .getLifeCycleController(systemRecoveryComponent);
            systemRecoveryLCC.stopFc();
        } catch (IllegalLifeCycleException e) {
            // Exceptions occuring during component stopping can be ignored
            System.out.println("WARNING: Unable to stop the Fractal component \"Recovery\"");
        }
		
	}

	/**
	 * Initialize the SSL properties if the corresponding parameters are set in
	 * the server.properties configuration file.
	 * 
	 * @throws IOException
	 */
	private void initializeSSL() throws IOException {

		if (!"".equals(SystemUtil.getTrustedkeystore())) {
			System.setProperty("javax.net.ssl.trustStore", SystemUtil
					.getTrustedkeystore());
		}

		if (!"".equals(SystemUtil.getTrustedkeystorePassword())) {
			System.setProperty("javax.net.ssl.trustStorePassword", SystemUtil
					.getTrustedkeystorePassword());
		}

		if (!"".equals(SystemUtil.getIdentitykeystore())) {
			System.setProperty("javax.net.ssl.keyStore", SystemUtil
					.getIdentitykeystore());
		}

		if (!"".equals(SystemUtil.getIdentitykeystorePassword())) {
			System.setProperty("javax.net.ssl.keyStorePassword", SystemUtil
					.getIdentitykeystorePassword());
		}
	}

	/**
	 * Set the Petals listener
	 * 
	 * @param petalsListener
	 */
	public void addStopListener(PetalsListener petalsListener) {
		this.petalsListener = petalsListener;
	}

	/**
	 * Remove the Petals listener
	 * 
	 */
	public void removeListener() {
		this.petalsListener = null;
	}

	/**
	 * Prints the Petals version on standard output. We connect to the JMX
	 * server of the server that is running with the server.properties entries.
	 * 
	 * @throws PetalsException
	 */
	public void getServerVersion() throws PetalsException {

		try {
			// get the jmxServer of this container
			MBeanServerConnection connection = mbeanHelper.retrieveJMXServer();

			// search the Admin mbean
			ObjectName result = mbeanHelper.retrieveServiceMBean(
					MBeanHelper.ADMIN_MBEAN, connection);

			// call Stop
			if (result != null) {
				String version = (String) connection.invoke(result,
						VERSION_ACTION, new Object[0], new String[0]);

				System.out.println(version);

			}
		} catch (Exception exception) {
			throw new PetalsException(exception);
		}
	}

	/**
	 * Initialize the PEtALS environment. Init Qname, JMX if necessarry,
	 * Monolog, Fractal and SSL.
	 * 
	 * @param jmxEmbedded
	 * 		<code>true</code> if a JMX server is already running in the PEtALS
	 * 		environment
	 */
	public void init(boolean jmxEmbedded) throws PetalsException {

		// set the compatible SUID for the Qname
		System.setProperty(
				"com.sun.xml.namespace.QName.useCompatibleSerialVersionUID",
				"1.0");

		if (!jmxEmbedded) {
			System.setProperty("javax.management.builder.initial",
					"mx4j.server.MX4JMBeanServerBuilder");
		}

		try {
			// initialize monolog
			this.initializeMonolog();

			// initialize Petals Fractal composite
			this.initializePetalsComposite();
		} catch (Exception exception) {
			// clean up the petals server
			throw new PetalsException("Problem while initializing Petals",
					exception);
		}
	}

	/**
	 * Start the Petals Server. Start the fractal components. Register the
	 * different component, then try to recover the Petals container.
	 * 
	 * @throws PetalsException
	 */
	public void start() throws PetalsException {

		try {
			// start Petals platform
			this.startPetalsComposite();

			// register the Petals server in the Petals admin.
			this.registerPetalsServer();

			// expose JBI MBean object in the JMX Server
			mbeanHelper.registerMBeans(petalsComposite);

			// recover all entities
			this.recoverSystem();

			this.startupDone();

			// add the lock file
			this.locker.lock();
		} catch (Exception exception) {
			// clean up the petals server
			try {
				this.startupFailure();
			} catch (Exception e) {
				// do nothing
			}

			if (petalsComposite != null) {
				try {
					fractalHelper.stopComponent(petalsComposite);
				} catch (Exception e) {
					// do nothing
				}
			}

			throw new PetalsException("Problem while starting Petals",
					exception);
		}

		try {
			this.initializeSSL();
		} catch (IOException e) {
			throw new PetalsException("SSL Properties cannot be correctly set!");
		}

		// notify the listener if there is one
		if (this.petalsListener != null) {
			this.petalsListener.onPetalsStarted();
		}
	}

	/**
	 * Trigger the stop or shutdown statement. We connect to the JMX server,
	 * locally or remotely, and invoke the statement.
	 * 
	 * @param shutdown
	 * 		shutdown if <code>, <code>stop</code> if false
	 * @param jmxEmbedded
	 * 		<code>true</code> if the JMX server is embedded to the PEtALS serverd
	 * 
	 * @throws PetalsException
	 */
	public void triggerStop(boolean shutdown, boolean jmxEmbedded)
			throws PetalsException {
		String action = (shutdown ? SHUTDOWN_ACTION : STOP_ACTION);

		try {
			if (jmxEmbedded) {
				this.triggerLocalStop(action);
			} else {
				this.triggerRemoteStop(action);
			}
		} catch (PetalsException e) {
			this.shutdownFailure();
			throw e;
		}

	}

	/**
	 * Invoke the given statement locally
	 * 
	 * @param action
	 * 		The action to perform
	 */
	public void triggerLocalStop(String action) throws PetalsException {
		MBeanServer jmxServer = null;
		ObjectName petalsAdminMBeanName = null;
		try {
			jmxServer = mbeanHelper.findLocalJMXServer(petalsComposite);
			petalsAdminMBeanName = new ObjectName(MBeanHelper.DOMAIN + ":name="
					+ MBeanHelper.PETALS_ADMIN + ",type=service");
			jmxServer.isRegistered(petalsAdminMBeanName);
		} catch (Exception e) {
			String msg = "Petals admin can not be accessed.";
			throw new PetalsException(msg, e);
		}

		try {
			// call the stop procedure and wait the thread stop to be processed
			jmxServer.invoke(petalsAdminMBeanName, action, new Object[0],
					new String[0]);
			this.petalsStopThread.join();
		} catch (Exception e) {
			throw new PetalsException(e);
		}
	}

	/**
	 * Invoke the given statement remotely
	 * 
	 * @param action
	 */
	public void triggerRemoteStop(String action) throws PetalsException {
		MBeanServerConnection connection = null;
		ObjectName result = null;

		try {
			// get the jmxServer of this container
			connection = mbeanHelper.retrieveJMXServer();

			// search the PetalsAdmin mbean
			result = mbeanHelper.retrieveServiceMBean(MBeanHelper.PETALS_ADMIN,
					connection);
		} catch (Exception e) {
			String msg = "Petals admin can not be accessed.";
			throw new PetalsException(msg, e);
		}

		if (result != null) {
			// call the stop procedure and wait the thread stop to be processed
			try {
				connection.invoke(result, action, new Object[0], new String[0]);
				this.petalsStopThread.join();
			} catch (Exception e) {
				throw new PetalsException(e);
			}
		}
	}

	/**
	 * Stop or Shutdown the Petals server.
	 * 
	 * @throws PetalsException
	 */
	public void stop() throws PetalsException {

		try {
			this.shutdownStart();
			this.stopPetalsComposite();
		} catch (Exception e) {
			this.shutdownFailure();
			throw new PetalsException(e);
		}

		this.locker.unlock();

		// notify the listener if there is one
		if (this.petalsListener != null) {
			this.petalsListener.onPetalsStopped();
		}

		// NOTE : unreachable part if the listener exists the system, case of
		// standalone

	}

	/**
	 * 
	 * @return
	 */
	public final boolean isLocked() {
		return this.locker.isLocked();
	}

	/**
	 * Set the local container state to started. The local container state is in
	 * unknow state until all is really started.
	 * 
	 * @throws Exception
	 */
	private void startupDone() throws Exception {
		Component network = fractalHelper.getRecursiveComponentByName(
				petalsContentController,
				FractalHelper.CONTAINERNETWORK_COMPONENT);

		if (network == null) {
			throw new NullPointerException(
					"The network fractal component is null");
		} else {
			NetworkService service = (NetworkService) network
					.getFcInterface("service");
			service
					.updateLocalContainerState(ContainerInformation.ContainerState.STARTED);
		}
	}

	/**
	 * Remove local container from information if its state is unknown. It meens
	 * that the container is not in good state and that there is an error during
	 * startup. If this operation is not done, the container can be binded in
	 * the container network but not really present;
	 * 
	 * @throws Exception
	 */
	private void startupFailure() throws Exception {
		LifeCycleController networkLc = fractalHelper
				.getLifeCycleControllerByName(petalsContentController,
						FractalHelper.CONTAINERNETWORK_COMPONENT);

		if (networkLc.getFcState() == LifeCycleController.STARTED) {
			Component network = fractalHelper.getRecursiveComponentByName(
					petalsContentController,
					FractalHelper.CONTAINERNETWORK_COMPONENT);

			if (network != null) {
				NetworkService service = (NetworkService) network
						.getFcInterface("service");
				ContainerInformation info = service
						.getLocalContainerInformation();
				if (info.getState() == ContainerInformation.ContainerState.UNKNOWN) {
					service.unregisterContainer(info.getName());
				}
			}
		}
	}

	/**
	 * Set the local container state to stopped. If the stopping is failing, the
	 * state of the container must changed to unknown
	 * 
	 * @throws Exception
	 */
	private void shutdownStart() throws Exception {
		Component network = fractalHelper.getRecursiveComponentByName(
				petalsContentController,
				FractalHelper.CONTAINERNETWORK_COMPONENT);

		if (network != null) {
			NetworkService service = (NetworkService) network
					.getFcInterface("service");
			service
					.updateLocalContainerState(ContainerInformation.ContainerState.STOPPED);
		}
	}

	/**
	 * Set the local container state to unknown if a failure occurs during the
	 * stop statement.
	 * 
	 * @throws Exception
	 */
	private void shutdownFailure() {
		LifeCycleController networkLc = fractalHelper
				.getLifeCycleControllerByName(petalsContentController,
						FractalHelper.CONTAINERNETWORK_COMPONENT);

		if (networkLc.getFcState() == LifeCycleController.STARTED) {
			Component network = fractalHelper.getRecursiveComponentByName(
					petalsContentController,
					FractalHelper.CONTAINERNETWORK_COMPONENT);

			try {
				if (network != null) {
					NetworkService service = (NetworkService) network
							.getFcInterface("service");
					ContainerInformation info = service
							.getLocalContainerInformation();
					if (info.getState() != ContainerInformation.ContainerState.UNKNOWN) {
						service
								.updateLocalContainerState(ContainerInformation.ContainerState.UNKNOWN);
					}
				}
			} catch (Exception e) {
				// do nothing
			}
		}
	}

}
