/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2008 OW2 Consortium, http://www.ow2.org/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.util;

import java.io.IOException;
import java.io.InputStream;
import java.util.Iterator;
import java.util.LinkedList;

/**
 * Initial developer(s): Christophe DENEUX
 * 
 * @author cdeneux - Capgemini Sud
 * @since 4.0
 * 
 */
public class InputStreamForker {

    private InputStream originalInputStream;

    private final LinkedList<byte[]> bufferListIs1;

    private final InputStream is1;

    private final LinkedList<byte[]> bufferListIs2;

    private final InputStream is2;

    /**
     * {@code true} if the {@code close()} method has been called on one of two
     * forked input stream
     */
    protected boolean inputStreamClosed = false;

    /**
     * {@code true} if the end of the original stream has been reached
     */
    protected boolean endStreamReached = false;

    public InputStreamForker(final InputStream originalInputStream) {
        this.originalInputStream = originalInputStream;

        this.bufferListIs1 = new LinkedList<byte[]>();
        this.bufferListIs2 = new LinkedList<byte[]>();

        this.is1 = new ForkedInputStream(this.bufferListIs1, this.bufferListIs2);
        this.is2 = new ForkedInputStream(this.bufferListIs2, this.bufferListIs1);

    }

    public InputStream getInputStreamOne() {
        return this.is1;
    }

    public InputStream getInputStreamTwo() {
        return this.is2;
    }

    protected class ForkedInputStream extends InputStream {

        final private LinkedList<byte[]> bufferList;

        /**
         * A reference on the other branch of inputstream.
         */
        final private LinkedList<byte[]> otherBufferList;

        public ForkedInputStream(final LinkedList<byte[]> bufferList,
                final LinkedList<byte[]> otherBufferList) {
            this.bufferList = bufferList;
            this.otherBufferList = otherBufferList;
        }

        @Override
        public int available() throws IOException {
            synchronized (InputStreamForker.this) {

                int available = 0;

                final Iterator<byte[]> iterator = this.bufferList.iterator();

                while (iterator.hasNext()) {
                    available += iterator.next().length;
                }

                if (!InputStreamForker.this.inputStreamClosed) {
                    available += InputStreamForker.this.originalInputStream.available();
                }

                return available;
            }
        }

        @Override
        public void close() throws IOException {
            synchronized (InputStreamForker.this) {
                InputStreamForker.this.inputStreamClosed = true;
                InputStreamForker.this.originalInputStream.close();
            }
        }

        @Override
        public boolean markSupported() {
            return false;
        }

        /**
         * <p>
         * Reads the next byte of data from the input stream. The value byte is
         * returned as an <code>int</code> in the range <code>0</code> to
         * <code>255</code>. If no byte is available because the end of the
         * stream has been reached, the value <code>-1</code> is returned. This
         * method blocks until input data is available, the end of the stream is
         * detected, or an exception is thrown.
         * </p>
         * <p>
         * The byte is read in the internal buffer. If no data is available in
         * the internal buffer, the byte is read from the original input stream.
         * </p>
         * 
         * @return the next byte of data, or <code>-1</code> if the end of the
         *  stream is reached.
         * @exception IOException
         *      if an I/O error occurs.
         */
        @Override
        public int read() throws IOException {
            byte[] buffer = new byte[1];
            int nbByteRead = this.read(buffer);
            if (nbByteRead != 1) {
                throw new IOException("No byte read.");
            }
            return buffer[0];
        }

        @Override
        public int read(final byte[] cbuf) throws IOException {
            return this.read(cbuf, 0, cbuf.length);
        }

        @Override
        public int read(final byte[] cbuf, final int off, final int len) throws IOException {
            return this.localRead(cbuf, off, len, false);
        }

        @Override
        public long skip(long n) throws IOException, IllegalArgumentException {
            if (n > Long.MAX_VALUE) {
                throw new IllegalArgumentException("Only value lesser " + Integer.MAX_VALUE
                        + "are accepted.");
            }
            return this.localRead(null, 0, (int) n, true);
        }

        private int localRead(final byte[] cbuf, final int off, final int len, final boolean skip)
                throws IOException {
            int remainingLen = len;
            int offset = off;
            int byteRead;

            synchronized (InputStreamForker.this) {
                // First we read from the internal buffers
                int nbBuffers = this.bufferList.size();
                while (nbBuffers > 0 && remainingLen != 0) {
                    final byte[] buffer = this.bufferList.getFirst();

                    if (remainingLen >= buffer.length) {
                        // Full copy needed
                        if (!skip) {
                            System.arraycopy(buffer, 0, cbuf, offset, buffer.length);
                        }

                        this.bufferList.removeFirst();

                        remainingLen -= buffer.length;
                        if (remainingLen == 0) {
                            // No more bytes must be read
                            return len;
                        }
                        offset += buffer.length;

                        nbBuffers--;
                    } else {
                        // Partial copy needed because the buffer len to read is
                        // reached
                        if (!skip) {
                            System.arraycopy(buffer, 0, cbuf, offset, remainingLen);
                        }

                        // Replace byte buffer by a buffer containing remaining
                        // bytes
                        final byte[] newBuffer = new byte[buffer.length - remainingLen];
                        System.arraycopy(buffer, remainingLen, newBuffer, 0, buffer.length
                                - remainingLen);
                        this.bufferList.removeFirst();
                        this.bufferList.addFirst(newBuffer);

                        return len;
                    }
                }

                // Next, we read missing bytes from the original input stream

                // Throw the exception only if the end of the original
                // stream has not been reached
                if (InputStreamForker.this.inputStreamClosed) {
                    if (InputStreamForker.this.endStreamReached) {
                        if (len != remainingLen) {
                            return len - remainingLen;
                        } else {
                            return -1;
                        }

                    }
                    throw new IOException(
                            "InputStream has been closed; cannot read from a closed InputStream.");
                }

                // Orginal input stream reading add buffer list management must
                // be synchronized.
                // Otherwise, inversions can occur.
                byte[] buffer = new byte[remainingLen];
                byteRead = InputStreamForker.this.originalInputStream.read(buffer, 0, remainingLen);

                if (byteRead != -1) {

                    // We add the read bytes into the other branch.
                    if (byteRead == remainingLen) {
                        // We can add the buffer as is because of its length is
                        // right.
                        this.otherBufferList.addLast(buffer);
                    } else {
                        final byte[] newBuffer = new byte[byteRead];
                        System.arraycopy(buffer, 0, newBuffer, 0, byteRead);
                        this.otherBufferList.addLast(newBuffer);
                    }

                    if (!skip) {
                        System.arraycopy(buffer, 0, cbuf, offset, remainingLen);
                    }

                    return len - remainingLen + byteRead;
                } else {
                    InputStreamForker.this.endStreamReached = true;
                    // return -1 if it is the end of the original buffer
                    if (len == remainingLen) {
                        return -1;
                    } else {
                        return len - remainingLen;
                    }
                }
            }
        }
    }
}
