/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: PropertyUtil.java 14:47:22 ddesjardins $
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.util;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;

import org.objectweb.petals.PetalsException;

/**
 * The property util is used to load and retrieve properties from file
 *
 * @author ddesjardins - eBMWebsourcing
 * @author chamerling - eBMWebsourcing
 */
public final class PropertyUtil {

    /**
     * Name of the server properties file
     */
    public static final String SERVER_PROPS = "/server.properties";

    public static final String UPDATED_PROPS = "/updated.properties";

    /**
     * Creates a new instance of {@link PropertyUtil}
     *
     */
    private PropertyUtil() {
    }

    /**
     * Retreive a property from a properties file
     *
     * @param propertiesFile
     *            name of the properties file (must be in the classpath)
     * @param key
     *            key of the property we want to retrieve
     * @return property associated with the key or null if not found
     * @throws PetalsException
     *             If a key is not defined in the property file or if any
     *             problems occurs while opening the file
     */
    public static String getProperty(String propertiesFile, String key)
            throws PetalsException {
        String value = null;
        try {
            Properties properties = loadPropertiesFile(propertiesFile);
            value = properties.getProperty(key);
        } catch (Exception e) {
            throw new PetalsException("Problem while trying to get property "
                    + key + " of " + propertiesFile, e);
        }
        return value;
    }

    /**
     * Load the specified properties file.
     *
     * @param propertiesFile
     *            name of the properties file (must be in the classpath)
     * @return a Properties object mapping given properties file
     * @throws IOException
     */
    public static Properties loadPropertiesFile(String propertiesFile) {
        Properties props = null;
        URL url = PropertyUtil.class.getResource(propertiesFile);
        props = new Properties();
        try {
            props.load(url.openStream());
        } catch (Exception e) {
            props = new Properties();
        }
        return props;
    }

    /**
     * Load the default properties file from the server.properties file
     *
     * @return
     * @throws IOException
     */
    public static Map<String, String> loadServerProperties() {
        return propertiesToMap(loadPropertiesFile(PropertyUtil.SERVER_PROPS));
    }

    /**
     *
     * @return
     */
    public static Map<String, String> loadUpdatedProperties() {
        return propertiesToMap(loadPropertiesFile(PropertyUtil.UPDATED_PROPS));
    }

    /**
     *
     * @param props
     * @return
     */
    protected static final Map<String, String> propertiesToMap(Properties props) {
        Map<String, String> map = new HashMap<String, String>();
        if (props != null) {
            Iterator<Object> it = props.keySet().iterator();
            while (it.hasNext()) {
                Object key = it.next();
                map.put(key.toString(), props.get(key).toString());
            }
        }
        return map;
    }

    /**
     *
     * @return
     * @throws IOException
     */
    public static Map<String, String> loadGlobalProperties() {
        Map<String, String> map = new HashMap<String, String>();
        map.putAll(loadServerProperties());
        // update only the keys that are not defined in the server properties file
        HashMap<String, String> updated = (HashMap<String, String>)loadUpdatedProperties();
        for(String key : updated.keySet()) {
            if (map.get(key) == null) {
                map.put(key, updated.get(key));
            }
        }
        
        return map;
    }

    /**
     * Retrieve a property from a properties file
     *
     * @param propertiesFile
     *            name of the properties file (must be in the classpath)
     * @param key
     *            key of the property we want to retrieve
     * @param defaultValue
     *            default value if no value is found with the given key
     * @return property associated with the key
     * @throws PetalsException
     *             If any problems occurs while opening the file
     */
    public static String getProperty(String propertiesFile, String key,
            String defaultValue) throws PetalsException {
        String value = null;
        try {
            Properties properties = loadPropertiesFile(propertiesFile);
            value = properties.getProperty(key, defaultValue);
        } catch (Exception e) {
            throw new PetalsException("Problem while trying to get property "
                    + key + " of " + propertiesFile, e);
        }
        return value;
    }

    /**
     * Save the properties to server.properties file
     *
     * @param props
     * @param propertiesFileName
     * @throws PetalsException
     */
    protected static final void saveProperties(Properties props,
            String fileResourceName) throws PetalsException {
        URL url = PropertyUtil.class.getResource(fileResourceName);
        try {
            File outFile = new File(url.toURI());
            saveProperties(props, outFile);
        } catch (URISyntaxException e) {
            throw new PetalsException(e);
        }
    }

    /**
     * Save the properties in a file.
     *
     * @param props
     * @param outputFile
     * @throws PetalsException
     */
    protected static final void saveProperties(Properties props, File outputFile)
            throws PetalsException {
        if (outputFile == null) {
            throw new PetalsException("Output file is null");
        }

        try {
            props.store(new FileOutputStream(outputFile), "PEtALS properties");
        } catch (FileNotFoundException e) {
            throw new PetalsException(e);
        } catch (IOException e) {
            throw new PetalsException(e);
        }
    }
}
