/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 Fossil E-Commerce, http://www.fossilec.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: ZipUtil.java 2948 2007-05-03 16:08:30Z chamerling $
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.util;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URI;
import java.util.Enumeration;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;

import org.apache.commons.io.IOUtils;
import org.objectweb.petals.PetalsException;

/**
 * Contains utilities methods in relation to Zip operations.
 * 
 * @author ddesjardins, ofabre - eBMWebsourcing
 */
public final class ZipUtil {

    /**
     * Do not construct.
     */
    private ZipUtil() {
        super();
    }

    /**
     * Create a temporary File from a given entry of a given ZipFile. You must
     * delete the temp file after using it.
     * 
     * @param zipFile
     *            the Zipped file to search in, must be non null
     * @param entryName
     *            the name of an entry in the Zip file, must be non null and non
     *            empty
     * @return a temp File read from the Zip entry, not <code>null</code>.
     * @throws PetalsException
     *             if entry can't be found or if an error occured during zip
     *             reading
     */
    public static File getEntryAsTemp(ZipFile zipFile, String entryName)
            throws PetalsException {
        ParameterCheckHelper.isNullParameter(zipFile, "zipFile");
        ParameterCheckHelper.isNullParameter(entryName, "entryName");
        ParameterCheckHelper.isEmptyParameter(entryName, "entryName");

        File tempFile = null;
        InputStream zipStream = null;
        try {
            ZipEntry theEntry = getZipEntry(zipFile, entryName);
            zipStream = zipFile.getInputStream(theEntry);
            tempFile = File.createTempFile("petals-", "-zip.tmp");
            NioUtil.copyStreamToFile(zipStream, tempFile);
        } catch (IOException e) {
            throw new PetalsException(e);
        } finally {
            try {
                if (zipStream != null) {
                    zipStream.close();
                }
            } catch (IOException ioe) {
                // do nothing
            }
        }
        return tempFile;
    }

    private static ZipEntry getZipEntry(ZipFile zipFile, String entryName)
            throws PetalsException {
        ZipEntry theEntry = null;
        if (entryName.indexOf("/") > -1) {
            theEntry = zipFile.getEntry(entryName);
            if (theEntry == null) {
                theEntry = zipFile.getEntry(entryName.replace("/", "\\"));
                if (theEntry == null) {
                    throw new PetalsException("Zip Entry not found: "
                            + entryName);
                }
            }
        } else {
            theEntry = zipFile.getEntry(entryName);
            if (theEntry == null) {
                theEntry = zipFile.getEntry(entryName.replace("\\", "/"));
                if (theEntry == null) {
                    throw new PetalsException("Zip Entry not found: "
                            + entryName);
                }
            }
        }
        return theEntry;
    }

    /**
     * Load a zip file from a URI
     * 
     * @param archiveURI
     *            uri of the zip file
     * @return zip file pointing on this URI
     * @throws NullPointerException
     *             if URI is null
     * @throws PetalsException
     *             if an error occured during zip file creation
     */
    public static ZipFile openZipFile(URI archiveURI) throws PetalsException {
        ZipFile zipArchive = null;
        String msg;
        try {
            File archiveFile = new File(archiveURI);
            zipArchive = new ZipFile(archiveFile);
        } catch (ZipException ze) {
            msg = "Error opening Zip file. Location: " + archiveURI.toString();
            throw new PetalsException(msg, ze);
        } catch (IOException ioe) {
            msg = "Unexpected I/O exception handling Zip file. Location : "
                    + archiveURI.toString();
            throw new PetalsException(msg, ioe);
        }
        return zipArchive;
    }

    private static void unzipEntry(ZipFile zipFile, ZipEntry entry, File destDir)
            throws IOException {
        InputStream in = null;
        OutputStream out = null;
        try {
            // Open the compressed entry
            in = zipFile.getInputStream(entry);

            // Open the output file
            out = new FileOutputStream(destDir);

            // Transfer bytes from the compressed file to the output file
            IOUtils.copy(in, out);

        } finally {
            // Close the file and stream
            if (in != null) {
                in.close();
            }
            if (out != null) {
                out.close();
            }
        }
    }

    /**
     * Unzip given zip file into the given destination
     * 
     * @param zipFile
     *            the File to unzip, must not be null
     * @param destDir
     *            the destination directory, can be null
     * @throws IOException
     *             if an I/O error occurred
     */
    public static void unzipArchive(File zipFile, File destDir)
            throws IOException {
        ParameterCheckHelper.isNullParameter(zipFile, "zipFile");

        ZipFile zip = new ZipFile(zipFile);
        try {
            Enumeration en = zip.entries();
            while (en.hasMoreElements()) {
                ZipEntry entry = (ZipEntry) en.nextElement();
                File file = null;
                if (destDir != null) {
                    file = new File(destDir, entry.getName());
                } else {
                    file = new File(entry.getName());
                }

                if (entry.isDirectory()) {
                    if (!file.mkdirs()) {
                        throw new IOException("Error creating directory: "
                                + file);
                    }
                } else {
                    File parent = file.getParentFile();
                    if (parent != null && !parent.exists()) {
                        if (!parent.mkdirs()) {
                            throw new IOException("Error creating directory: "
                                    + parent);
                        }
                    }
                    unzipEntry(zip, entry, file);
                }
            }
        } finally {
            zip.close();
        }
    }
}
