/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2009 OW2, http://www.ow2.org/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: $
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.extensions.transporter.nio;

import java.io.IOException;
import java.util.Map;
import java.util.Queue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentLinkedQueue;

import javax.jbi.messaging.MessageExchange;

import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.fraclet.annotation.annotations.FractalComponent;
import org.objectweb.fractal.fraclet.annotation.annotations.Interface;
import org.objectweb.fractal.fraclet.annotation.annotations.LifeCycle;
import org.objectweb.fractal.fraclet.annotation.annotations.Monolog;
import org.objectweb.fractal.fraclet.annotation.annotations.Provides;
import org.objectweb.fractal.fraclet.annotation.annotations.Requires;
import org.objectweb.fractal.fraclet.annotation.annotations.type.LifeCycleType;
import org.objectweb.petals.communication.jndi.JNDIService;
import org.objectweb.petals.jbi.messaging.exchange.MessageExchangeImpl;
import org.objectweb.petals.jbi.messaging.servicedesc.AbstractEndpoint;
import org.objectweb.petals.jbi.messaging.transport.TransportException;
import org.objectweb.petals.jbi.messaging.transport.TransportProtocol;
import org.objectweb.petals.jbi.messaging.transport.TransportProtocolListener;
import org.objectweb.petals.jbi.messaging.transport.util.TransporterUtil;
import org.objectweb.petals.kernel.configuration.LocalConfiguration;
import org.objectweb.petals.util.LoggingUtil;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.ow2.petals.extensions.transporter.nio.client.NioClientObjectPool;
import org.ow2.petals.extensions.transporter.nio.selector.NioClient;
import org.ow2.petals.extensions.transporter.nio.server.NioReceiver;


/**
 * NIO Implementation of the Transporter
 * 
 * @author Christophe DENEUX (Capgemini Sud)
 */
@FractalComponent
@Provides(interfaces = {
		@Interface(name = "service", signature = TransportProtocol.class),
		@Interface(name = NioReceiver.NIO_MSG_RECEIVER_ITF, signature = NioReceiver.class) })
public class NioTransporterImpl implements TransportProtocol, NioReceiver {

    /**
     * Fractal Logger
     */
    @Monolog(name = "logger")
    protected Logger logger;

    /**
	 * Interface used to access configuration
	 */
	@Requires(name = CONFIGURATION_ITF, signature = LocalConfiguration.class)
	protected LocalConfiguration configurationService;

	/**
	 * Interface used to dispatch messages.
	 */
	@Requires(name = DISPATCHER_ITF, signature = TransportProtocolListener.class)
	protected TransportProtocolListener dispatcher;

	/**
	 * Interface used to access registry
	 */
	@Requires(name = JNDI_ITF, signature = JNDIService.class)
	protected JNDIService jndiService;
	
	/**
	 * All connections to remote containers are pooled into a keyed-pool. The
	 * key is the name of the remote container.
	 */
    @Requires(name = NioClientObjectPool.NIO_CLIENT_OBJECT_ITF, signature = NioClientObjectPool.class)
	protected NioClientObjectPool clientConnections;

	/**
     * Logger Wrapper
     */
    protected LoggingUtil log;

    /**
     * Map used to store incoming messages for poll() call
     */
    private Map<String, Queue<MessageExchangeImpl>> incomingExchanges;
    
    /**
	 * The maximum size of the blocking queue containing message exchanges that
	 * wait to be processed by component.
	 */
	private int queueSize = Parameters.DEFAULT_QUEUE_SIZE;
	
	/**
	 * Timeout to send a message exchange.
	 */
	private long sendTimeout;

	/**
     * @see TransportProtocol#addDestination(String)
     */
    public void addDestination(final String componentName) throws TransportException {
        log.call("Destination : " + componentName);

        if (incomingExchanges.containsKey(componentName)) {
            throw new TransportException("The destination " + componentName
                    + " already exist");
        } else {
            incomingExchanges.put(componentName, 
                    new ConcurrentLinkedQueue<MessageExchangeImpl>());
        }
    }

    /**
     * @see TransportProtocol#receive(java.lang.String)
     */
    public MessageExchangeImpl receive(final String componentName)
            throws TransportException {
        log.start();

        final Queue<MessageExchangeImpl> queue = this.incomingExchanges
                .get(componentName);
        
        // As queue is a ConcurrentLinkedQueue, no synchronization is needed
        final MessageExchangeImpl responseExchange = queue.poll();

        log.end();
        return responseExchange;
    }
    
    public void onMessage(final Object object) {
    	this.log.start();
    	if (object instanceof MessageExchangeImpl) {
			final MessageExchangeImpl messageExchange = (MessageExchangeImpl)object;
			
    		if (TransporterUtil.getSyncMode(messageExchange, true)) {
    			if (this.logger.isLoggable(BasicLevel.DEBUG)) {
        			log.debug("The message" + messageExchange.getExchangeId()
                                + " is a synchronized response for the component "
                                + getComponentName(messageExchange));
                }
    		    dispatcher.syncMessageReceived(messageExchange);
    		} else {
    		    if (this.logger.isLoggable(BasicLevel.DEBUG)) {
    		        log.debug("The message " + messageExchange.getExchangeId()
                                + " is not a synchronized response");
    		    }

    		    final String componentName = NioTransporterImpl.getComponentName(messageExchange);

    			final Queue<MessageExchangeImpl> exchangeQueue = this.incomingExchanges
						.get(componentName);

    			if (exchangeQueue == null) {
                    log
                        .warning("The component '"
                                + componentName
                                + "' is not ready to receive message, or is unknown. The message exchange is rejected");
                } else if (exchangeQueue.size() == queueSize) {
					log
						.warning("The Nio Transporter has reached its maximum capacity for the target component '"
								+ componentName
								+ "'. The message exchange is rejected");
				} else {
					// As queue is a ConcurrentLinkedQueue, no synchronization is needed
					exchangeQueue.add(messageExchange);

					// Warn the dispatcher
					dispatcher.messageAvailable(this, componentName);
				}
    		}
    	}
		else {
			// This case doesn't have to exist.
			this.log.warning("Unexpected received data. Data are discarded.");
		}
    	this.log.end();
    }
        
    private static final String getComponentName(final MessageExchangeImpl jbiMsg) {
		String componentName = null;
		if (jbiMsg.getRole().equals(MessageExchange.Role.CONSUMER)) {
			componentName = (jbiMsg.getConsumerEndpoint()).getComponentName();
		} else if (jbiMsg.getRole().equals(MessageExchange.Role.PROVIDER)) {
			componentName = ((AbstractEndpoint) jbiMsg.getEndpoint())
					.getComponentName();
		}

		return componentName;
	}

    /**
     * @see TransportProtocol#removeDestination(java.lang.String)
     */
    public void removeDestination(String componentName)
            throws TransportException {
        log.start("Destination : " + componentName);

        if (this.incomingExchanges.remove(componentName) == null) {
            log.error("The destination " + componentName + " does not exist");
        }
    }

    /**
     * This method does following action :
     * <ol>
     * <li>Serialize the message exchange.</li>
     * <li>Get the destination of the message.</li>
     * <li>Add it to the message.</li>
     * <li>Send it to the transport agent.</li>
     * </ol>
     * 
     * @see TransportProtocol#send(MessageExchangeImpl,
     *      java.lang.String, java.lang.String, long)
     */
    public void send(MessageExchangeImpl messageExchange, String componentName,
            String containerName, long timeToLive) throws TransportException {
        log.start("Send message (via Nio) to container : " + containerName
                + " / component : " + componentName);

        // clean unused messages
        messageExchange.cleanMessages();
        
        try {
        	final NioClient clientConnection = this.clientConnections.borrowObject(containerName);
        	
        	try {
	    		clientConnection.send(messageExchange, this.sendTimeout);
	    		
	    		// In case of error during send, the clientConnection is automatically invalidate.

        	} finally  {
				this.clientConnections.returnObject(containerName, clientConnection);
			}
        	
		} catch (final IOException e) {
			throw new TransportException(e);
		} catch (InterruptedException e) {
			throw new TransportException(e);
		} catch (Exception e) {
			throw new TransportException(e);
		} finally {
			log.end();
		}
    }
    
    /**
     * Init variables with Fractal attributes value and create/bind the topic
     * for this container.
     * 
     * @throws NoSuchInterfaceException
     * 
     * @throws Exception
     */
    @LifeCycle(on = LifeCycleType.START)
    public void start() throws IllegalLifeCycleException,
            NoSuchInterfaceException {
        this.log = new LoggingUtil(logger);
        this.log.start();
        
        // Maps initialization
        this.incomingExchanges = new ConcurrentHashMap<String, Queue<MessageExchangeImpl>>(10);
        
        this.sendTimeout = Long.parseLong(this.configurationService
				.getValue(Parameters.PROP_NAME_SEND_TIMEOUT,
						Parameters.DEFAULT_SEND_TIMEOUT));
        
    }

    /**
     * Method called when the fractal component is stopped
     * 
     * @throws Exception
     */
    @LifeCycle(on = LifeCycleType.STOP)
    public void stop() throws Exception {
        this.log.start();
        
        // NOP
        
        this.log.end();
    }
    /* (non-Javadoc)
     * @see org.objectweb.petals.PetalsLifeCycle#shutdown()
     */
    public void shutdown() throws Exception {
        // NOP
    }

}
