/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2009 OW2, http://www.ow2.org/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: $
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.extensions.transporter.nio.selector;

import java.io.IOException;
import java.io.OutputStream;
import java.net.SocketException;
import java.nio.ByteBuffer;
import java.nio.channels.SelectionKey;
import java.nio.channels.SocketChannel;

/**
 * The socket channel context of an input socket. It contains following
 * information:
 * <ul>
 * <li>the socket channel,</li>
 * <li>its direct NIO byte buffer,</li>
 * <li>its pending read,</li>
 * <li>a server callback to process read bytes.</li>
 * </ul>
 * 
 * @author Christophe DENEUX (Capgemini Sud)
 */
public class InputSocketChannelContext extends AbstractSocketChannelContext {

    /**
     * Output stream to used to push data that have been read from the TCP
     * socket to the NIO server agent.
     */
    private final OutputStream bytesStream;

    /**
     * Deserializer associated to this context
     */
    private final PipedDeserializer deserializer;

    /**
     * The NIO server agent to invoke to process read bytes.
     */
    private NioServer nioServer;

    final byte[] headerBuffer = new byte[INTEGER_SIZE_BYTE];

    final ByteBuffer integerBuffer = ByteBuffer.allocateDirect(INTEGER_SIZE_BYTE);

    int headerRemainingBytes = 0;

    boolean isHeaderCompletlyRead = true;

    /**
     * Constructor to use on server socket channel
     * 
     * @param socketChannel
     *            The socket channel on which a context must be created.
     * @param onReadCallback
     *            The callback of the server agent to use when reading bytes.
     * @throws SocketException
     */
    public InputSocketChannelContext(final SocketChannel socketChannel,
            final PipedDeserializer deserializer, final NioServer nioServer) throws SocketException {
        // TODO: The size of the direct buffer to use to read should be
        // configured instead of using the SO_RCVBUF size.
        super(socketChannel, socketChannel.socket().getReceiveBufferSize());

        this.deserializer = deserializer;
        this.bytesStream = deserializer.getOutputStream();
        this.nioServer = nioServer;
    }

    public void read(final SelectionKey key) throws IOException {

        // We prepare the buffer to read bytes
        this.directBuffer.clear();

        // 1/ We read available bytes
        int remainingBytesInBuffer;
        try {
            remainingBytesInBuffer = this.socketChannel.read(this.directBuffer);
        } catch (final IOException e) {
            // The client has forcibly closed its connection. We free local
            // resources: we close the socket and cancel the associated selector
            // key.
            this.closeConnection(key);
            return;
        }

        if (remainingBytesInBuffer == -1) {
            // The client has ended the connection. We free local resources: we
            // close the socket and cancel the associated selector key.
            this.closeConnection(key);
        } else {
            // Prepare the buffer for reading bytes
            this.directBuffer.flip();

            do {
                if (!this.isHeaderCompletlyRead) {
                    // We read the remaining part of the header (externalized
                    // message exchange size)
                    if (remainingBytesInBuffer >= this.headerRemainingBytes) {
                        // The header can be completed
                        this.isHeaderCompletlyRead = true;
                        this.directBuffer.get(this.headerBuffer, INTEGER_SIZE_BYTE
                                - this.headerRemainingBytes, this.headerRemainingBytes);

                        // At this point we have a byte array containing the
                        // header. To prevent bits ordering we must do a small
                        // work.
                        this.integerBuffer.clear();
                        this.integerBuffer.put(this.headerBuffer);
                        this.integerBuffer.flip();
                        this.msgExRemainingBytes = this.integerBuffer.getInt();

                        remainingBytesInBuffer -= this.headerRemainingBytes;

                    } else {
                        // The header is not completed again
                        this.directBuffer.get(this.headerBuffer, INTEGER_SIZE_BYTE
                                - this.headerRemainingBytes, remainingBytesInBuffer);
                        this.headerRemainingBytes -= remainingBytesInBuffer;

                        // At this point, no more byte can be read
                        remainingBytesInBuffer = 0;
                    }
                }
                // As if "isHeaderCompletlyRead" is false, "isIOPending" isn't
                // false, we can use the clause "else" to prevent the evaluation
                // of the following expression
                else if (!this.isIOPending) {
                    // No externalized message exchange is reading, we start to
                    // read an new one.

                    // 2/ We read header (externalized message exchange size)
                    if (remainingBytesInBuffer >= INTEGER_SIZE_BYTE) {
                        // 2a/ First, the header containing the size of the
                        // externalized message exchange.
                        this.msgExRemainingBytes = this.directBuffer.getInt();
                        remainingBytesInBuffer -= INTEGER_SIZE_BYTE;

                        this.isHeaderCompletlyRead = true;
                    } else {
                        // The header is not completely read
                        this.isHeaderCompletlyRead = false;

                        this.headerRemainingBytes = INTEGER_SIZE_BYTE - remainingBytesInBuffer;
                        this.directBuffer.get(this.headerBuffer, 0, remainingBytesInBuffer);

                        // At this point, no more byte can be read
                        remainingBytesInBuffer = 0;

                    }
                    this.isIOPending = true;
                }

                if (remainingBytesInBuffer > 0) {
                    // 3/ Next, we read remaining bytes of the current
                    // externalized message exchange.
                    final byte[] buf = new byte[Math.min(this.msgExRemainingBytes,
                            remainingBytesInBuffer)];
                    this.directBuffer.get(buf);
                    this.bytesStream.write(buf);

                    // 4/ If the read of the message exchange is complete, no
                    // pending read is in progress. Otherwise the externalized
                    // message exchange read continue.
                    if (buf.length == this.msgExRemainingBytes) {
                        this.endMsgExRead();
                    } else {
                        this.msgExRemainingBytes -= buf.length;
                    }

                    remainingBytesInBuffer -= buf.length;
                }

            } while (remainingBytesInBuffer > 0);
        }
    }

    /**
     * Close connection.
     * 
     * @param key
     * @throws IOException
     */
    private void closeConnection(final SelectionKey key) throws IOException {
        System.err.println("The client has forcibly closed its connection.");
        this.socketChannel.close();
        key.cancel();
        this.nioServer.onClose(this.socketChannel);
    }

    /**
     * End reading of message Exchange.
     */
    private final void endMsgExRead() {
        this.isIOPending = false;
        this.nioServer.onRead(this.deserializer);
    }

}
